@@ -21,9 +21,11 @@
 #include <grub/command.h>
 #include <grub/i18n.h>
 #include <grub/memory.h>
+#include <grub/mm.h>
 
 GRUB_MOD_LICENSE ("GPLv3+");
 
+#ifndef GRUB_MACHINE_EMU
 static const char *names[] =
   {
     [GRUB_MEMORY_AVAILABLE] = N_("available RAM"),
@@ -35,30 +37,35 @@ static const char *names[] =
        is required to save accross hibernations.  */
     [GRUB_MEMORY_NVS] = N_("ACPI non-volatile storage RAM"),
     [GRUB_MEMORY_BADRAM] = N_("faulty RAM (BadRAM)"),
-    [GRUB_MEMORY_CODE] = N_("RAM holding firmware code"),
-    [GRUB_MEMORY_HOLE] = N_("Address range not associated with RAM")
+    [GRUB_MEMORY_PERSISTENT] = N_("persistent RAM"),
+    [GRUB_MEMORY_PERSISTENT_LEGACY] = N_("persistent RAM (legacy)"),
+    [GRUB_MEMORY_COREBOOT_TABLES] = N_("RAM holding coreboot tables"),
+    [GRUB_MEMORY_CODE] = N_("RAM holding firmware code")
   };
 
+/* Helper for grub_cmd_lsmmap.  */
+static int
+lsmmap_hook (grub_uint64_t addr, grub_uint64_t size, grub_memory_type_t type,
+	     void *data __attribute__ ((unused)))
+{
+  if (type < (int) ARRAY_SIZE (names) && type >= 0 && names[type])
+    grub_printf_ (N_("base_addr = 0x%llx, length = 0x%llx, %s\n"),
+		  (long long) addr, (long long) size, _(names[type]));
+  else
+    grub_printf_ (N_("base_addr = 0x%llx, length = 0x%llx, type = 0x%x\n"),
+		  (long long) addr, (long long) size, type);
+  return 0;
+}
+#endif
+
 static grub_err_t
 grub_cmd_lsmmap (grub_command_t cmd __attribute__ ((unused)),
 		 int argc __attribute__ ((unused)),
 		 char **args __attribute__ ((unused)))
 
 {
-  auto int NESTED_FUNC_ATTR hook (grub_uint64_t, grub_uint64_t, grub_memory_type_t);
-  int NESTED_FUNC_ATTR hook (grub_uint64_t addr, grub_uint64_t size, 
-			     grub_memory_type_t type)
-    {
-      if (type < ARRAY_SIZE (names) && names[type])
-	grub_printf_ (N_("base_addr = 0x%llx, length = 0x%llx, %s\n"),
-		      (long long) addr, (long long) size, _(names[type]));
-      else
-	grub_printf_ (N_("base_addr = 0x%llx, length = 0x%llx, type = 0x%x\n"),
-		      (long long) addr, (long long) size, type);
-      return 0;
-    }
 #ifndef GRUB_MACHINE_EMU
-  grub_machine_mmap_iterate (hook);
+  grub_machine_mmap_iterate (lsmmap_hook, NULL);
 #endif
 
   return 0;
