@@ -17,6 +17,8 @@
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#define grub_fshelp_node grub_ntfs_file 
+
 #include <grub/file.h>
 #include <grub/mm.h>
 #include <grub/misc.h>
@@ -30,64 +32,80 @@ GRUB_MOD_LICENSE ("GPLv3+");
 
 static grub_dl_t my_mod;
 
-ntfscomp_func_t grub_ntfscomp_func;
+#define grub_fshelp_node grub_ntfs_file 
+
+static inline grub_uint16_t
+u16at (void *ptr, grub_size_t ofs)
+{
+  return grub_le_to_cpu16 (grub_get_unaligned16 ((char *) ptr + ofs));
+}
+
+static inline grub_uint32_t
+u32at (void *ptr, grub_size_t ofs)
+{
+  return grub_le_to_cpu32 (grub_get_unaligned32 ((char *) ptr + ofs));
+}
+
+static inline grub_uint64_t
+u64at (void *ptr, grub_size_t ofs)
+{
+  return grub_le_to_cpu64 (grub_get_unaligned64 ((char *) ptr + ofs));
+}
+
+grub_ntfscomp_func_t grub_ntfscomp_func;
 
 static grub_err_t
-fixup (struct grub_ntfs_data *data, char *buf, int len, char *magic)
+fixup (grub_uint8_t *buf, grub_size_t len, const grub_uint8_t *magic)
 {
-  int ss;
-  char *pu;
+  grub_uint16_t ss;
+  grub_uint8_t *pu;
   grub_uint16_t us;
 
+  COMPILE_TIME_ASSERT ((1 << GRUB_NTFS_BLK_SHR) == GRUB_DISK_SECTOR_SIZE);
+
   if (grub_memcmp (buf, magic, 4))
     return grub_error (GRUB_ERR_BAD_FS, "%s label not found", magic);
 
   ss = u16at (buf, 6) - 1;
-  if (ss * (int) data->blocksize != len * GRUB_DISK_SECTOR_SIZE)
-    return grub_error (GRUB_ERR_BAD_FS, "size not match",
-		       ss * (int) data->blocksize,
-		       len * GRUB_DISK_SECTOR_SIZE);
+  if (ss != len)
+    return grub_error (GRUB_ERR_BAD_FS, "size not match");
   pu = buf + u16at (buf, 4);
   us = u16at (pu, 0);
   buf -= 2;
   while (ss > 0)
     {
-      buf += data->blocksize;
+      buf += GRUB_DISK_SECTOR_SIZE;
       pu += 2;
       if (u16at (buf, 0) != us)
 	return grub_error (GRUB_ERR_BAD_FS, "fixup signature not match");
-      v16at (buf, 0) = v16at (pu, 0);
+      buf[0] = pu[0];
+      buf[1] = pu[1];
       ss--;
     }
 
   return 0;
 }
 
-static grub_err_t read_mft (struct grub_ntfs_data *data, char *buf,
-			    grub_uint32_t mftno);
-static grub_err_t read_attr (struct grub_ntfs_attr *at, char *dest,
+static grub_err_t read_mft (struct grub_ntfs_data *data, grub_uint8_t *buf,
+			    grub_uint64_t mftno);
+static grub_err_t read_attr (struct grub_ntfs_attr *at, grub_uint8_t *dest,
 			     grub_disk_addr_t ofs, grub_size_t len,
 			     int cached,
-			     void
-			     NESTED_FUNC_ATTR (*read_hook) (grub_disk_addr_t
-							    sector,
-							    unsigned offset,
-							    unsigned length));
+			     grub_disk_read_hook_t read_hook,
+			     void *read_hook_data);
 
-static grub_err_t read_data (struct grub_ntfs_attr *at, char *pa, char *dest,
+static grub_err_t read_data (struct grub_ntfs_attr *at, grub_uint8_t *pa,
+			     grub_uint8_t *dest,
 			     grub_disk_addr_t ofs, grub_size_t len,
 			     int cached,
-			     void
-			     NESTED_FUNC_ATTR (*read_hook) (grub_disk_addr_t
-							    sector,
-							    unsigned offset,
-							    unsigned length));
+			     grub_disk_read_hook_t read_hook,
+			     void *read_hook_data);
 
 static void
 init_attr (struct grub_ntfs_attr *at, struct grub_ntfs_file *mft)
 {
   at->mft = mft;
-  at->flags = (mft == &mft->data->mmft) ? AF_MMFT : 0;
+  at->flags = (mft == &mft->data->mmft) ? GRUB_NTFS_AF_MMFT : 0;
   at->attr_nxt = mft->buf + u16at (mft->buf, 0x14);
   at->attr_end = at->emft_buf = at->edat_buf = at->sbuf = NULL;
 }
@@ -100,34 +118,33 @@ free_attr (struct grub_ntfs_attr *at)
   grub_free (at->sbuf);
 }
 
-static char *
-find_attr (struct grub_ntfs_attr *at, unsigned char attr)
+static grub_uint8_t *
+find_attr (struct grub_ntfs_attr *at, grub_uint8_t attr)
 {
-  if (at->flags & AF_ALST)
+  if (at->flags & GRUB_NTFS_AF_ALST)
     {
     retry:
       while (at->attr_nxt < at->attr_end)
 	{
 	  at->attr_cur = at->attr_nxt;
 	  at->attr_nxt += u16at (at->attr_cur, 4);
-	  if (((unsigned char) *at->attr_cur == attr) || (attr == 0))
+	  if ((*at->attr_cur == attr) || (attr == 0))
 	    {
-	      char *new_pos;
+	      grub_uint8_t *new_pos;
 
-	      if (at->flags & AF_MMFT)
+	      if (at->flags & GRUB_NTFS_AF_MMFT)
 		{
 		  if ((grub_disk_read
-		       (at->mft->data->disk, v32at (at->attr_cur, 0x10), 0,
+		       (at->mft->data->disk, u32at (at->attr_cur, 0x10), 0,
 			512, at->emft_buf))
 		      ||
 		      (grub_disk_read
-		       (at->mft->data->disk, v32at (at->attr_cur, 0x14), 0,
+		       (at->mft->data->disk, u32at (at->attr_cur, 0x14), 0,
 			512, at->emft_buf + 512)))
 		    return NULL;
 
-		  if (fixup
-		      (at->mft->data, at->emft_buf, at->mft->data->mft_size,
-		       "FILE"))
+		  if (fixup (at->emft_buf, at->mft->data->mft_size,
+			     (const grub_uint8_t *) "FILE"))
 		    return NULL;
 		}
 	      else
@@ -138,10 +155,9 @@ find_attr (struct grub_ntfs_attr *at, unsigned char attr)
 		}
 
 	      new_pos = &at->emft_buf[u16at (at->emft_buf, 0x14)];
-	      while ((unsigned char) *new_pos != 0xFF)
+	      while (*new_pos != 0xFF)
 		{
-		  if (((unsigned char) *new_pos ==
-		       (unsigned char) *at->attr_cur)
+		  if ((*new_pos == *at->attr_cur)
 		      && (u16at (new_pos, 0xE) == u16at (at->attr_cur, 0x18)))
 		    {
 		      return new_pos;
@@ -157,27 +173,27 @@ find_attr (struct grub_ntfs_attr *at, unsigned char attr)
       return NULL;
     }
   at->attr_cur = at->attr_nxt;
-  while ((unsigned char) *at->attr_cur != 0xFF)
+  while (*at->attr_cur != 0xFF)
     {
       at->attr_nxt += u16at (at->attr_cur, 4);
-      if ((unsigned char) *at->attr_cur == AT_ATTRIBUTE_LIST)
+      if (*at->attr_cur == GRUB_NTFS_AT_ATTRIBUTE_LIST)
 	at->attr_end = at->attr_cur;
-      if (((unsigned char) *at->attr_cur == attr) || (attr == 0))
+      if ((*at->attr_cur == attr) || (attr == 0))
 	return at->attr_cur;
       at->attr_cur = at->attr_nxt;
     }
   if (at->attr_end)
     {
-      char *pa;
+      grub_uint8_t *pa;
 
-      at->emft_buf = grub_malloc (at->mft->data->mft_size << BLK_SHR);
+      at->emft_buf = grub_malloc (at->mft->data->mft_size << GRUB_NTFS_BLK_SHR);
       if (at->emft_buf == NULL)
 	return NULL;
 
       pa = at->attr_end;
       if (pa[8])
 	{
-          int n;
+          grub_uint32_t n;
 
           n = ((u32at (pa, 0x30) + GRUB_DISK_SECTOR_SIZE - 1)
                & (~(GRUB_DISK_SECTOR_SIZE - 1)));
@@ -185,7 +201,7 @@ find_attr (struct grub_ntfs_attr *at, unsigned char attr)
 	  at->edat_buf = grub_malloc (n);
 	  if (!at->edat_buf)
 	    return NULL;
-	  if (read_data (at, pa, at->edat_buf, 0, n, 0, 0))
+	  if (read_data (at, pa, at->edat_buf, 0, n, 0, 0, 0))
 	    {
 	      grub_error (GRUB_ERR_BAD_FS,
 			  "fail to read non-resident attribute list");
@@ -199,59 +215,64 @@ find_attr (struct grub_ntfs_attr *at, unsigned char attr)
 	  at->attr_nxt = at->attr_end + u16at (pa, 0x14);
 	  at->attr_end = at->attr_end + u32at (pa, 4);
 	}
-      at->flags |= AF_ALST;
+      at->flags |= GRUB_NTFS_AF_ALST;
       while (at->attr_nxt < at->attr_end)
 	{
-	  if (((unsigned char) *at->attr_nxt == attr) || (attr == 0))
+	  if ((*at->attr_nxt == attr) || (attr == 0))
 	    break;
 	  at->attr_nxt += u16at (at->attr_nxt, 4);
 	}
       if (at->attr_nxt >= at->attr_end)
 	return NULL;
 
-      if ((at->flags & AF_MMFT) && (attr == AT_DATA))
+      if ((at->flags & GRUB_NTFS_AF_MMFT) && (attr == GRUB_NTFS_AT_DATA))
 	{
-	  at->flags |= AF_GPOS;
+	  at->flags |= GRUB_NTFS_AF_GPOS;
 	  at->attr_cur = at->attr_nxt;
 	  pa = at->attr_cur;
-	  v32at (pa, 0x10) = at->mft->data->mft_start;
-	  v32at (pa, 0x14) = at->mft->data->mft_start + 1;
+	  grub_set_unaligned32 ((char *) pa + 0x10,
+				grub_cpu_to_le32 (at->mft->data->mft_start));
+	  grub_set_unaligned32 ((char *) pa + 0x14,
+				grub_cpu_to_le32 (at->mft->data->mft_start
+						  + 1));
 	  pa = at->attr_nxt + u16at (pa, 4);
 	  while (pa < at->attr_end)
 	    {
-	      if ((unsigned char) *pa != attr)
+	      if (*pa != attr)
 		break;
 	      if (read_attr
 		  (at, pa + 0x10,
-		   u32at (pa, 0x10) * (at->mft->data->mft_size << BLK_SHR),
-		   at->mft->data->mft_size << BLK_SHR, 0, 0))
+		   u32at (pa, 0x10) * (at->mft->data->mft_size << GRUB_NTFS_BLK_SHR),
+		   at->mft->data->mft_size << GRUB_NTFS_BLK_SHR, 0, 0, 0))
 		return NULL;
 	      pa += u16at (pa, 4);
 	    }
 	  at->attr_nxt = at->attr_cur;
-	  at->flags &= ~AF_GPOS;
+	  at->flags &= ~GRUB_NTFS_AF_GPOS;
 	}
       goto retry;
     }
   return NULL;
 }
 
-static char *
+static grub_uint8_t *
 locate_attr (struct grub_ntfs_attr *at, struct grub_ntfs_file *mft,
-	     unsigned char attr)
+	     grub_uint8_t attr)
 {
-  char *pa;
+  grub_uint8_t *pa;
 
   init_attr (at, mft);
-  if ((pa = find_attr (at, attr)) == NULL)
+  pa = find_attr (at, attr);
+  if (pa == NULL)
     return NULL;
-  if ((at->flags & AF_ALST) == 0)
+  if ((at->flags & GRUB_NTFS_AF_ALST) == 0)
     {
       while (1)
 	{
-	  if ((pa = find_attr (at, attr)) == NULL)
+	  pa = find_attr (at, attr);
+	  if (pa == NULL)
 	    break;
-	  if (at->flags & AF_ALST)
+	  if (at->flags & GRUB_NTFS_AF_ALST)
 	    return pa;
 	}
       grub_errno = GRUB_ERR_NONE;
@@ -262,49 +283,40 @@ locate_attr (struct grub_ntfs_attr *at, struct grub_ntfs_file *mft,
   return pa;
 }
 
-static char *
-read_run_data (char *run, int nn, grub_disk_addr_t * val, int sig)
+static grub_disk_addr_t
+read_run_data (const grub_uint8_t *run, int nn, int sig)
 {
-  grub_disk_addr_t r, v;
+  grub_uint64_t r = 0;
 
-  r = 0;
-  v = 1;
+  if (sig && nn && (run[nn - 1] & 0x80))
+    r = -1;
 
-  while (nn--)
-    {
-      r += v * (*(unsigned char *) (run++));
-      v <<= 8;
-    }
+  grub_memcpy (&r, run, nn);
 
-  if ((sig) && (r & (v >> 1)))
-    r -= v;
-
-  *val = r;
-  return run;
+  return grub_le_to_cpu64 (r);
 }
 
 grub_err_t
 grub_ntfs_read_run_list (struct grub_ntfs_rlst * ctx)
 {
-  int c1, c2;
+  grub_uint8_t c1, c2;
   grub_disk_addr_t val;
-  char *run;
+  grub_uint8_t *run;
 
   run = ctx->cur_run;
 retry:
-  c1 = ((unsigned char) (*run) & 0xF);
-  c2 = ((unsigned char) (*run) >> 4);
+  c1 = ((*run) & 0x7);
+  c2 = ((*run) >> 4) & 0x7;
+  run++;
   if (!c1)
     {
-      if ((ctx->attr) && (ctx->attr->flags & AF_ALST))
+      if ((ctx->attr) && (ctx->attr->flags & GRUB_NTFS_AF_ALST))
 	{
-	  void NESTED_FUNC_ATTR (*save_hook) (grub_disk_addr_t sector,
-					      unsigned offset,
-					      unsigned length);
+	  grub_disk_read_hook_t save_hook;
 
 	  save_hook = ctx->comp.disk->read_hook;
 	  ctx->comp.disk->read_hook = 0;
-	  run = find_attr (ctx->attr, (unsigned char) *ctx->attr->attr_cur);
+	  run = find_attr (ctx->attr, *ctx->attr->attr_cur);
 	  ctx->comp.disk->read_hook = save_hook;
 	  if (run)
 	    {
@@ -319,15 +331,16 @@ retry:
 	}
       return grub_error (GRUB_ERR_BAD_FS, "run list overflown");
     }
-  run = read_run_data (run + 1, c1, &val, 0);	/* length of current VCN */
   ctx->curr_vcn = ctx->next_vcn;
-  ctx->next_vcn += val;
-  run = read_run_data (run, c2, &val, 1);	/* offset to previous LCN */
+  ctx->next_vcn += read_run_data (run, c1, 0);	/* length of current VCN */
+  run += c1;
+  val = read_run_data (run, c2, 1);	/* offset to previous LCN */
+  run += c2;
   ctx->curr_lcn += val;
   if (val == 0)
-    ctx->flags |= RF_BLNK;
+    ctx->flags |= GRUB_NTFS_RF_BLNK;
   else
-    ctx->flags &= ~RF_BLNK;
+    ctx->flags &= ~GRUB_NTFS_RF_BLNK;
   ctx->cur_run = run;
   return 0;
 }
@@ -345,18 +358,15 @@ grub_ntfs_read_block (grub_fshelp_node_t node, grub_disk_addr_t block)
       return ctx->curr_lcn;
     }
   else
-    return (ctx->flags & RF_BLNK) ? 0 : (block -
+    return (ctx->flags & GRUB_NTFS_RF_BLNK) ? 0 : (block -
 					 ctx->curr_vcn + ctx->curr_lcn);
 }
 
 static grub_err_t
-read_data (struct grub_ntfs_attr *at, char *pa, char *dest,
+read_data (struct grub_ntfs_attr *at, grub_uint8_t *pa, grub_uint8_t *dest,
 	   grub_disk_addr_t ofs, grub_size_t len, int cached,
-	   void NESTED_FUNC_ATTR (*read_hook) (grub_disk_addr_t sector,
-					       unsigned offset,
-					       unsigned length))
+	   grub_disk_read_hook_t read_hook, void *read_hook_data)
 {
-  grub_disk_addr_t vcn;
   struct grub_ntfs_rlst cc, *ctx;
 
   if (len == 0)
@@ -365,9 +375,12 @@ read_data (struct grub_ntfs_attr *at, char *pa, char *dest,
   grub_memset (&cc, 0, sizeof (cc));
   ctx = &cc;
   ctx->attr = at;
-  ctx->comp.spc = at->mft->data->spc;
+  ctx->comp.log_spc = at->mft->data->log_spc;
   ctx->comp.disk = at->mft->data->disk;
 
+  if (read_hook == grub_file_progress_hook)
+    ctx->file = read_hook_data;
+
   if (pa[8] == 0)
     {
       if (ofs + len > u32at (pa, 0x10))
@@ -376,121 +389,88 @@ read_data (struct grub_ntfs_attr *at, char *pa, char *dest,
       return 0;
     }
 
-  if (u16at (pa, 0xC) & FLAG_COMPRESSED)
-    ctx->flags |= RF_COMP;
-  else
-    ctx->flags &= ~RF_COMP;
   ctx->cur_run = pa + u16at (pa, 0x20);
 
-  if (ctx->flags & RF_COMP)
+  ctx->next_vcn = u32at (pa, 0x10);
+  ctx->curr_lcn = 0;
+
+  if ((pa[0xC] & GRUB_NTFS_FLAG_COMPRESSED)
+      && !(at->flags & GRUB_NTFS_AF_GPOS))
     {
       if (!cached)
 	return grub_error (GRUB_ERR_BAD_FS, "attribute can\'t be compressed");
 
-      if (at->sbuf)
-	{
-	  if ((ofs & (~(COM_LEN - 1))) == at->save_pos)
-	    {
-	      grub_disk_addr_t n;
-
-	      n = COM_LEN - (ofs - at->save_pos);
-	      if (n > len)
-		n = len;
-
-	      grub_memcpy (dest, at->sbuf + ofs - at->save_pos, n);
-	      if (n == len)
-		return 0;
-
-	      dest += n;
-	      len -= n;
-	      ofs += n;
-	    }
-	}
-      else
-	{
-	  at->sbuf = grub_malloc (COM_LEN);
-	  if (at->sbuf == NULL)
-	    return grub_errno;
-	  at->save_pos = 1;
-	}
-
-      vcn = ctx->target_vcn = (ofs >> COM_LOG_LEN) * (COM_SEC / ctx->comp.spc);
-      ctx->target_vcn &= ~0xF;
+      return (grub_ntfscomp_func) ? grub_ntfscomp_func (dest, ofs, len, ctx)
+	: grub_error (GRUB_ERR_BAD_FS, N_("module `%s' isn't loaded"),
+		      "ntfscomp");
     }
-  else
-    vcn = ctx->target_vcn = grub_divmod64 (ofs >> BLK_SHR, ctx->comp.spc, 0);
 
-  ctx->next_vcn = u32at (pa, 0x10);
-  ctx->curr_lcn = 0;
+  ctx->target_vcn = ofs >> (GRUB_NTFS_BLK_SHR + ctx->comp.log_spc);
   while (ctx->next_vcn <= ctx->target_vcn)
     {
       if (grub_ntfs_read_run_list (ctx))
 	return grub_errno;
     }
 
-  if (at->flags & AF_GPOS)
+  if (at->flags & GRUB_NTFS_AF_GPOS)
     {
       grub_disk_addr_t st0, st1;
-      grub_uint32_t m;
+      grub_uint64_t m;
 
-      grub_divmod64 (ofs >> BLK_SHR, ctx->comp.spc, &m);
+      m = (ofs >> GRUB_NTFS_BLK_SHR) & ((1 << ctx->comp.log_spc) - 1);
 
       st0 =
-	(ctx->target_vcn - ctx->curr_vcn + ctx->curr_lcn) * ctx->comp.spc + m;
+	((ctx->target_vcn - ctx->curr_vcn + ctx->curr_lcn) << ctx->comp.log_spc) + m;
       st1 = st0 + 1;
       if (st1 ==
-	  (ctx->next_vcn - ctx->curr_vcn + ctx->curr_lcn) * ctx->comp.spc)
+	  (ctx->next_vcn - ctx->curr_vcn + ctx->curr_lcn) << ctx->comp.log_spc)
 	{
 	  if (grub_ntfs_read_run_list (ctx))
 	    return grub_errno;
-	  st1 = ctx->curr_lcn * ctx->comp.spc;
+	  st1 = ctx->curr_lcn << ctx->comp.log_spc;
 	}
-      v32at (dest, 0) = st0;
-      v32at (dest, 4) = st1;
+      grub_set_unaligned32 (dest, grub_cpu_to_le32 (st0));
+      grub_set_unaligned32 (dest + 4, grub_cpu_to_le32 (st1));
       return 0;
     }
 
-  if (!(ctx->flags & RF_COMP))
-    {
-      unsigned int pow;
-
-      if (!grub_fshelp_log2blksize (ctx->comp.spc, &pow))
-	grub_fshelp_read_file (ctx->comp.disk, (grub_fshelp_node_t) ctx,
-			       read_hook, ofs, len, dest,
-			       grub_ntfs_read_block, ofs + len, pow);
-      return grub_errno;
-    }
-
-  return (grub_ntfscomp_func) ? grub_ntfscomp_func (at, dest, ofs, len, ctx,
-						    vcn) :
-    grub_error (GRUB_ERR_BAD_FS, "ntfscomp module not loaded");
+  grub_fshelp_read_file (ctx->comp.disk, (grub_fshelp_node_t) ctx,
+			 read_hook, read_hook_data, ofs, len,
+			 (char *) dest,
+			 grub_ntfs_read_block, ofs + len,
+			 ctx->comp.log_spc, 0);
+  return grub_errno;
 }
 
 static grub_err_t
-read_attr (struct grub_ntfs_attr *at, char *dest, grub_disk_addr_t ofs,
+read_attr (struct grub_ntfs_attr *at, grub_uint8_t *dest, grub_disk_addr_t ofs,
 	   grub_size_t len, int cached,
-	   void NESTED_FUNC_ATTR (*read_hook) (grub_disk_addr_t sector,
-					       unsigned offset,
-					       unsigned length))
+	   grub_disk_read_hook_t read_hook, void *read_hook_data)
 {
-  char *save_cur;
-  unsigned char attr;
-  char *pp;
+  grub_uint8_t *save_cur;
+  grub_uint8_t attr;
+  grub_uint8_t *pp;
   grub_err_t ret;
 
   save_cur = at->attr_cur;
   at->attr_nxt = at->attr_cur;
-  attr = (unsigned char) *at->attr_nxt;
-  if (at->flags & AF_ALST)
+  attr = *at->attr_nxt;
+  if (at->flags & GRUB_NTFS_AF_ALST)
     {
-      char *pa;
+      grub_uint8_t *pa;
       grub_disk_addr_t vcn;
 
-      vcn = grub_divmod64 (ofs, at->mft->data->spc << BLK_SHR, 0);
+      /* If compression is possible make sure that we include possible
+	 compressed block size.  */
+      if (GRUB_NTFS_LOG_COM_SEC >= at->mft->data->log_spc)
+	vcn = ((ofs >> GRUB_NTFS_COM_LOG_LEN)
+	       << (GRUB_NTFS_LOG_COM_SEC - at->mft->data->log_spc)) & ~0xFULL;
+      else
+	vcn = ofs >> (at->mft->data->log_spc + GRUB_NTFS_BLK_SHR);
       pa = at->attr_nxt + u16at (at->attr_nxt, 4);
       while (pa < at->attr_end)
 	{
-	  if ((unsigned char) *pa != attr)
+	  if (*pa != attr)
 	    break;
 	  if (u32at (pa, 8) > vcn)
 	    break;
@@ -500,7 +480,8 @@ read_attr (struct grub_ntfs_attr *at, char *dest, grub_disk_addr_t ofs,
     }
   pp = find_attr (at, attr);
   if (pp)
-    ret = read_data (at, pp, dest, ofs, len, cached, read_hook);
+    ret = read_data (at, pp, dest, ofs, len, cached,
+		     read_hook, read_hook_data);
   else
     ret =
       (grub_errno) ? grub_errno : grub_error (GRUB_ERR_BAD_FS,
@@ -510,23 +491,23 @@ read_attr (struct grub_ntfs_attr *at, char *dest, grub_disk_addr_t ofs,
 }
 
 static grub_err_t
-read_mft (struct grub_ntfs_data *data, char *buf, grub_uint32_t mftno)
+read_mft (struct grub_ntfs_data *data, grub_uint8_t *buf, grub_uint64_t mftno)
 {
   if (read_attr
-      (&data->mmft.attr, buf, mftno * ((grub_disk_addr_t) data->mft_size << BLK_SHR),
-       data->mft_size << BLK_SHR, 0, 0))
-    return grub_error (GRUB_ERR_BAD_FS, "read MFT 0x%X fails", mftno);
-  return fixup (data, buf, data->mft_size, "FILE");
+      (&data->mmft.attr, buf, mftno * ((grub_disk_addr_t) data->mft_size << GRUB_NTFS_BLK_SHR),
+       data->mft_size << GRUB_NTFS_BLK_SHR, 0, 0, 0))
+    return grub_error (GRUB_ERR_BAD_FS, "read MFT 0x%llx fails", (unsigned long long) mftno);
+  return fixup (buf, data->mft_size, (const grub_uint8_t *) "FILE");
 }
 
 static grub_err_t
-init_file (struct grub_ntfs_file *mft, grub_uint32_t mftno)
+init_file (struct grub_ntfs_file *mft, grub_uint64_t mftno)
 {
   unsigned short flag;
 
   mft->inode_read = 1;
 
-  mft->buf = grub_malloc (mft->data->mft_size << BLK_SHR);
+  mft->buf = grub_malloc (mft->data->mft_size << GRUB_NTFS_BLK_SHR);
   if (mft->buf == NULL)
     return grub_errno;
 
@@ -535,22 +516,24 @@ init_file (struct grub_ntfs_file *mft, grub_uint32_t mftno)
 
   flag = u16at (mft->buf, 0x16);
   if ((flag & 1) == 0)
-    return grub_error (GRUB_ERR_BAD_FS, "MFT 0x%X is not in use", mftno);
+    return grub_error (GRUB_ERR_BAD_FS, "MFT 0x%llx is not in use",
+		       (unsigned long long) mftno);
 
   if ((flag & 2) == 0)
     {
-      char *pa;
+      grub_uint8_t *pa;
 
-      pa = locate_attr (&mft->attr, mft, AT_DATA);
+      pa = locate_attr (&mft->attr, mft, GRUB_NTFS_AT_DATA);
       if (pa == NULL)
-	return grub_error (GRUB_ERR_BAD_FS, "no $DATA in MFT 0x%X", mftno);
+	return grub_error (GRUB_ERR_BAD_FS, "no $DATA in MFT 0x%llx",
+			   (unsigned long long) mftno);
 
       if (!pa[8])
 	mft->size = u32at (pa, 0x10);
       else
 	mft->size = u64at (pa, 0x30);
 
-      if ((mft->attr.flags & AF_ALST) == 0)
+      if ((mft->attr.flags & GRUB_NTFS_AF_ALST) == 0)
 	mft->attr.attr_end = 0;	/*  Don't jump to attribute list */
     }
   else
@@ -566,25 +549,45 @@ free_file (struct grub_ntfs_file *mft)
   grub_free (mft->buf);
 }
 
-static int
-list_file (struct grub_ntfs_file *diro, char *pos,
-	   int NESTED_FUNC_ATTR
-	   (*hook) (const char *filename,
-		    enum grub_fshelp_filetype filetype,
-		    grub_fshelp_node_t node))
+static char *
+get_utf8 (grub_uint8_t *in, grub_size_t len)
 {
-  char *np;
+  grub_uint8_t *buf;
+  grub_uint16_t *tmp;
+  grub_size_t i;
+
+  buf = grub_malloc (len * GRUB_MAX_UTF8_PER_UTF16 + 1);
+  tmp = grub_malloc (len * sizeof (tmp[0]));
+  if (!buf || !tmp)
+    {
+      grub_free (buf);
+      grub_free (tmp);
+      return NULL;
+    }
+  for (i = 0; i < len; i++)
+    tmp[i] = grub_le_to_cpu16 (grub_get_unaligned16 (in + 2 * i));
+  *grub_utf16_to_utf8 (buf, tmp, len) = '\0';
+  grub_free (tmp);
+  return (char *) buf;
+}
+
+static int
+list_file (struct grub_ntfs_file *diro, grub_uint8_t *pos,
+	   grub_fshelp_iterate_dir_hook_t hook, void *hook_data)
+{
+  grub_uint8_t *np;
   int ns;
 
   while (1)
     {
-      char *ustr, namespace;
+      grub_uint8_t namespace;
+      char *ustr;
 
       if (pos[0xC] & 2)		/* end signature */
 	break;
 
       np = pos + 0x50;
-      ns = (unsigned char) *(np++);
+      ns = *(np++);
       namespace = *(np++);
 
       /*
@@ -595,34 +598,34 @@ list_file (struct grub_ntfs_file *diro, char *pos,
 	{
 	  enum grub_fshelp_filetype type;
 	  struct grub_ntfs_file *fdiro;
+	  grub_uint32_t attr;
 
-	  if (u16at (pos, 4))
-	    {
-	      grub_error (GRUB_ERR_BAD_FS, "64-bit MFT number");
-	      return 0;
-	    }
-
-	  type =
-	    (u32at (pos, 0x48) & ATTR_DIRECTORY) ? GRUB_FSHELP_DIR :
-	    GRUB_FSHELP_REG;
+	  attr = u32at (pos, 0x48);
+	  if (attr & GRUB_NTFS_ATTR_REPARSE)
+	    type = GRUB_FSHELP_SYMLINK;
+	  else if (attr & GRUB_NTFS_ATTR_DIRECTORY)
+	    type = GRUB_FSHELP_DIR;
+	  else
+	    type = GRUB_FSHELP_REG;
 
 	  fdiro = grub_zalloc (sizeof (struct grub_ntfs_file));
 	  if (!fdiro)
 	    return 0;
 
 	  fdiro->data = diro->data;
-	  fdiro->ino = u32at (pos, 0);
+	  fdiro->ino = u64at (pos, 0) & 0xffffffffffffULL;
+	  fdiro->mtime = u64at (pos, 0x20);
 
-	  ustr = grub_malloc (ns * 4 + 1);
+	  ustr = get_utf8 (np, ns);
 	  if (ustr == NULL)
-	    return 0;
-	  *grub_utf16_to_utf8 ((grub_uint8_t *) ustr, (grub_uint16_t *) np,
-			       ns) = '\0';
-
+	    {
+	      grub_free (fdiro);
+	      return 0;
+	    }
           if (namespace)
             type |= GRUB_FSHELP_CASE_INSENSITIVE;
 
-	  if (hook (ustr, type, fdiro))
+	  if (hook (ustr, type, fdiro, hook_data))
 	    {
 	      grub_free (ustr);
 	      return 1;
@@ -635,16 +638,105 @@ list_file (struct grub_ntfs_file *diro, char *pos,
   return 0;
 }
 
+struct symlink_descriptor
+{
+  grub_uint32_t type;
+  grub_uint32_t total_len;
+  grub_uint16_t off1;
+  grub_uint16_t len1;
+  grub_uint16_t off2;
+  grub_uint16_t len2;
+} GRUB_PACKED;
+
+static char *
+grub_ntfs_read_symlink (grub_fshelp_node_t node)
+{
+  struct grub_ntfs_file *mft;
+  struct symlink_descriptor symdesc;
+  grub_err_t err;
+  grub_uint8_t *buf16;
+  char *buf, *end;
+  grub_size_t len;
+  grub_uint8_t *pa;
+  grub_size_t off;
+
+  mft = (struct grub_ntfs_file *) node;
+
+  mft->buf = grub_malloc (mft->data->mft_size << GRUB_NTFS_BLK_SHR);
+  if (mft->buf == NULL)
+    return NULL;
+
+  if (read_mft (mft->data, mft->buf, mft->ino))
+    return NULL;
+
+  pa = locate_attr (&mft->attr, mft, GRUB_NTFS_AT_SYMLINK);
+  if (pa == NULL)
+    {
+      grub_error (GRUB_ERR_BAD_FS, "no $SYMLINK in MFT 0x%llx",
+		  (unsigned long long) mft->ino);
+      return NULL;
+    }
+
+  err = read_attr (&mft->attr, (grub_uint8_t *) &symdesc, 0,
+		   sizeof (struct symlink_descriptor), 1, 0, 0);
+  if (err)
+    return NULL;
+
+  switch (grub_cpu_to_le32 (symdesc.type))
+    {
+    case 0xa000000c:
+      off = (sizeof (struct symlink_descriptor) + 4
+	     + grub_cpu_to_le32 (symdesc.off1));
+      len = grub_cpu_to_le32 (symdesc.len1);
+      break;
+    case 0xa0000003:
+      off = (sizeof (struct symlink_descriptor)
+	     + grub_cpu_to_le32 (symdesc.off1));
+      len = grub_cpu_to_le32 (symdesc.len1);
+      break;
+    default:
+      grub_error (GRUB_ERR_BAD_FS, "symlink type invalid (%x)",
+		  grub_cpu_to_le32 (symdesc.type));
+      return NULL;
+    }
+
+  buf16 = grub_malloc (len);
+  if (!buf16)
+    return NULL;
+
+  err = read_attr (&mft->attr, buf16, off, len, 1, 0, 0);
+  if (err)
+    return NULL;
+
+  buf = get_utf8 (buf16, len / 2);
+  if (!buf)
+    {
+      grub_free (buf16);
+      return NULL;
+    }
+  grub_free (buf16);
+
+  for (end = buf; *end; end++)
+    if (*end == '\\')
+      *end = '/';
+
+  /* Split the sequence to avoid GCC thinking that this is a trigraph.  */
+  if (grub_memcmp (buf, "/?" "?/", 4) == 0 && buf[5] == ':' && buf[6] == '/'
+      && grub_isalpha (buf[4]))
+    {
+      grub_memmove (buf, buf + 6, end - buf + 1 - 6);
+      end -= 6; 
+    }
+  return buf;
+}
+
 static int
 grub_ntfs_iterate_dir (grub_fshelp_node_t dir,
-		       int NESTED_FUNC_ATTR
-		       (*hook) (const char *filename,
-				enum grub_fshelp_filetype filetype,
-				grub_fshelp_node_t node))
+		       grub_fshelp_iterate_dir_hook_t hook, void *hook_data)
 {
-  unsigned char *bitmap;
+  grub_uint8_t *bitmap;
   struct grub_ntfs_attr attr, *at;
-  char *cur_pos, *indx, *bmp;
+  grub_uint8_t *cur_pos, *indx, *bmp;
   int ret = 0;
   grub_size_t bitmap_len;
   struct grub_ntfs_file *mft;
@@ -664,7 +756,8 @@ grub_ntfs_iterate_dir (grub_fshelp_node_t dir,
   init_attr (at, mft);
   while (1)
     {
-      if ((cur_pos = find_attr (at, AT_INDEX_ROOT)) == NULL)
+      cur_pos = find_attr (at, GRUB_NTFS_AT_INDEX_ROOT);
+      if (cur_pos == NULL)
 	{
 	  grub_error (GRUB_ERR_BAD_FS, "no $INDEX_ROOT");
 	  goto done;
@@ -682,7 +775,7 @@ grub_ntfs_iterate_dir (grub_fshelp_node_t dir,
     }
 
   cur_pos += 0x10;		/* Skip index root */
-  ret = list_file (mft, cur_pos + u16at (cur_pos, 0), hook);
+  ret = list_file (mft, cur_pos + u16at (cur_pos, 0), hook, hook_data);
   if (ret)
     goto done;
 
@@ -690,11 +783,11 @@ grub_ntfs_iterate_dir (grub_fshelp_node_t dir,
   bitmap_len = 0;
   free_attr (at);
   init_attr (at, mft);
-  while ((cur_pos = find_attr (at, AT_BITMAP)) != NULL)
+  while ((cur_pos = find_attr (at, GRUB_NTFS_AT_BITMAP)) != NULL)
     {
       int ofs;
 
-      ofs = (unsigned char) cur_pos[0xA];
+      ofs = cur_pos[0xA];
       /* Namelen=4, Name="$I30" */
       if ((cur_pos[9] == 4) &&
 	  (u32at (cur_pos, ofs) == 0x490024) &&
@@ -711,12 +804,12 @@ grub_ntfs_iterate_dir (grub_fshelp_node_t dir,
 
 	  if (is_resident)
 	    {
-              grub_memcpy (bmp, (char *) (cur_pos + u16at (cur_pos, 0x14)),
+              grub_memcpy (bmp, cur_pos + u16at (cur_pos, 0x14),
                            bitmap_len);
 	    }
           else
             {
-              if (read_data (at, cur_pos, bmp, 0, bitmap_len, 0, 0))
+              if (read_data (at, cur_pos, bmp, 0, bitmap_len, 0, 0, 0))
                 {
                   grub_error (GRUB_ERR_BAD_FS,
                               "fails to read non-resident $BITMAP");
@@ -725,13 +818,13 @@ grub_ntfs_iterate_dir (grub_fshelp_node_t dir,
               bitmap_len = u32at (cur_pos, 0x30);
             }
 
-          bitmap = (unsigned char *) bmp;
+          bitmap = bmp;
 	  break;
 	}
     }
 
   free_attr (at);
-  cur_pos = locate_attr (at, mft, AT_INDEX_ALLOCATION);
+  cur_pos = locate_attr (at, mft, GRUB_NTFS_AT_INDEX_ALLOCATION);
   while (cur_pos != NULL)
     {
       /* Non-resident, Namelen=4, Offset=0x40, Flags=0, Name="$I30" */
@@ -739,7 +832,7 @@ grub_ntfs_iterate_dir (grub_fshelp_node_t dir,
 	  (u32at (cur_pos, 0x40) == 0x490024) &&
 	  (u32at (cur_pos, 0x44) == 0x300033))
 	break;
-      cur_pos = find_attr (at, AT_INDEX_ALLOCATION);
+      cur_pos = find_attr (at, GRUB_NTFS_AT_INDEX_ALLOCATION);
     }
 
   if ((!cur_pos) && (bitmap))
@@ -750,9 +843,10 @@ grub_ntfs_iterate_dir (grub_fshelp_node_t dir,
 
   if (bitmap)
     {
-      grub_disk_addr_t v, i;
+      grub_disk_addr_t i;
+      grub_uint8_t v;
 
-      indx = grub_malloc (mft->data->idx_size << BLK_SHR);
+      indx = grub_malloc (mft->data->idx_size << GRUB_NTFS_BLK_SHR);
       if (indx == NULL)
 	goto done;
 
@@ -762,16 +856,18 @@ grub_ntfs_iterate_dir (grub_fshelp_node_t dir,
 	  if (*bitmap & v)
 	    {
 	      if ((read_attr
-		   (at, indx, i * (mft->data->idx_size << BLK_SHR),
-		    (mft->data->idx_size << BLK_SHR), 0, 0))
-		  || (fixup (mft->data, indx, mft->data->idx_size, "INDX")))
+		   (at, indx, i * (mft->data->idx_size << GRUB_NTFS_BLK_SHR),
+		    (mft->data->idx_size << GRUB_NTFS_BLK_SHR), 0, 0, 0))
+		  || (fixup (indx, mft->data->idx_size,
+			     (const grub_uint8_t *) "INDX")))
 		goto done;
-	      ret = list_file (mft, &indx[0x18 + u16at (indx, 0x18)], hook);
+	      ret = list_file (mft, &indx[0x18 + u16at (indx, 0x18)],
+			       hook, hook_data);
 	      if (ret)
 		goto done;
 	    }
 	  v <<= 1;
-	  if (v >= 0x100)
+	  if (!v)
 	    {
 	      v = 1;
 	      bitmap++;
@@ -792,6 +888,7 @@ grub_ntfs_mount (grub_disk_t disk)
 {
   struct grub_ntfs_bpb bpb;
   struct grub_ntfs_data *data = 0;
+  grub_uint32_t spc;
 
   if (!disk)
     goto fail;
@@ -806,47 +903,61 @@ grub_ntfs_mount (grub_disk_t disk)
   if (grub_disk_read (disk, 0, 0, sizeof (bpb), &bpb))
     goto fail;
 
-  if (grub_memcmp ((char *) &bpb.oem_name, "NTFS", 4))
+  if (grub_memcmp ((char *) &bpb.oem_name, "NTFS", 4) != 0
+      || bpb.sectors_per_cluster == 0
+      || (bpb.sectors_per_cluster & (bpb.sectors_per_cluster - 1)) != 0
+      || bpb.bytes_per_sector == 0
+      || (bpb.bytes_per_sector & (bpb.bytes_per_sector - 1)) != 0)
     goto fail;
 
-  data->blocksize = grub_le_to_cpu16 (bpb.bytes_per_sector);
-  data->spc = bpb.sectors_per_cluster * (data->blocksize >> BLK_SHR);
+  spc = (((grub_uint32_t) bpb.sectors_per_cluster
+	  * (grub_uint32_t) grub_le_to_cpu16 (bpb.bytes_per_sector))
+	 >> GRUB_NTFS_BLK_SHR);
+  if (spc == 0)
+    goto fail;
+
+  for (data->log_spc = 0; (1U << data->log_spc) < spc; data->log_spc++);
 
   if (bpb.clusters_per_mft > 0)
-    data->mft_size = data->spc * bpb.clusters_per_mft;
+    data->mft_size = ((grub_disk_addr_t) bpb.clusters_per_mft) << data->log_spc;
+  else if (-bpb.clusters_per_mft < GRUB_NTFS_BLK_SHR || -bpb.clusters_per_mft >= 31)
+    goto fail;
   else
-    data->mft_size = 1 << (-bpb.clusters_per_mft - BLK_SHR);
+    data->mft_size = 1ULL << (-bpb.clusters_per_mft - GRUB_NTFS_BLK_SHR);
 
   if (bpb.clusters_per_index > 0)
-    data->idx_size = data->spc * bpb.clusters_per_index;
+    data->idx_size = (((grub_disk_addr_t) bpb.clusters_per_index)
+		      << data->log_spc);
+  else if (-bpb.clusters_per_index < GRUB_NTFS_BLK_SHR || -bpb.clusters_per_index >= 31)
+    goto fail;
   else
-    data->idx_size = 1 << (-bpb.clusters_per_index - BLK_SHR);
+    data->idx_size = 1ULL << (-bpb.clusters_per_index - GRUB_NTFS_BLK_SHR);
 
-  data->mft_start = grub_le_to_cpu64 (bpb.mft_lcn) * data->spc;
+  data->mft_start = grub_le_to_cpu64 (bpb.mft_lcn) << data->log_spc;
 
-  if ((data->mft_size > MAX_MFT) || (data->idx_size > MAX_IDX))
+  if ((data->mft_size > GRUB_NTFS_MAX_MFT) || (data->idx_size > GRUB_NTFS_MAX_IDX))
     goto fail;
 
   data->mmft.data = data;
   data->cmft.data = data;
 
-  data->mmft.buf = grub_malloc (data->mft_size << BLK_SHR);
+  data->mmft.buf = grub_malloc (data->mft_size << GRUB_NTFS_BLK_SHR);
   if (!data->mmft.buf)
     goto fail;
 
   if (grub_disk_read
-      (disk, data->mft_start, 0, data->mft_size << BLK_SHR, data->mmft.buf))
+      (disk, data->mft_start, 0, data->mft_size << GRUB_NTFS_BLK_SHR, data->mmft.buf))
     goto fail;
 
   data->uuid = grub_le_to_cpu64 (bpb.num_serial);
 
-  if (fixup (data, data->mmft.buf, data->mft_size, "FILE"))
+  if (fixup (data->mmft.buf, data->mft_size, (const grub_uint8_t *) "FILE"))
     goto fail;
 
-  if (!locate_attr (&data->mmft.attr, &data->mmft, AT_DATA))
+  if (!locate_attr (&data->mmft.attr, &data->mmft, GRUB_NTFS_AT_DATA))
     goto fail;
 
-  if (init_file (&data->cmft, FILE_ROOT))
+  if (init_file (&data->cmft, GRUB_NTFS_FILE_ROOT))
     goto fail;
 
   return data;
@@ -863,29 +974,39 @@ fail:
   return 0;
 }
 
+/* Context for grub_ntfs_dir.  */
+struct grub_ntfs_dir_ctx
+{
+  grub_fs_dir_hook_t hook;
+  void *hook_data;
+};
+
+/* Helper for grub_ntfs_dir.  */
+static int
+grub_ntfs_dir_iter (const char *filename, enum grub_fshelp_filetype filetype,
+		    grub_fshelp_node_t node, void *data)
+{
+  struct grub_ntfs_dir_ctx *ctx = data;
+  struct grub_dirhook_info info;
+
+  grub_memset (&info, 0, sizeof (info));
+  info.dir = ((filetype & GRUB_FSHELP_TYPE_MASK) == GRUB_FSHELP_DIR);
+  info.mtimeset = 1;
+  info.mtime = grub_divmod64 (node->mtime, 10000000, 0) 
+    - 86400ULL * 365 * (1970 - 1601)
+    - 86400ULL * ((1970 - 1601) / 4) + 86400ULL * ((1970 - 1601) / 100);
+  grub_free (node);
+  return ctx->hook (filename, &info, ctx->hook_data);
+}
+
 static grub_err_t
 grub_ntfs_dir (grub_device_t device, const char *path,
-	       int (*hook) (const char *filename,
-			    const struct grub_dirhook_info *info))
+	       grub_fs_dir_hook_t hook, void *hook_data)
 {
+  struct grub_ntfs_dir_ctx ctx = { hook, hook_data };
   struct grub_ntfs_data *data = 0;
   struct grub_fshelp_node *fdiro = 0;
 
-  auto int NESTED_FUNC_ATTR iterate (const char *filename,
-				     enum grub_fshelp_filetype filetype,
-				     grub_fshelp_node_t node);
-
-  int NESTED_FUNC_ATTR iterate (const char *filename,
-				enum grub_fshelp_filetype filetype,
-				grub_fshelp_node_t node)
-  {
-      struct grub_dirhook_info info;
-      grub_memset (&info, 0, sizeof (info));
-      info.dir = ((filetype & GRUB_FSHELP_TYPE_MASK) == GRUB_FSHELP_DIR);
-      grub_free (node);
-      return hook (filename, &info);
-  }
-
   grub_dl_ref (my_mod);
 
   data = grub_ntfs_mount (device->disk);
@@ -893,12 +1014,12 @@ grub_ntfs_dir (grub_device_t device, const char *path,
     goto fail;
 
   grub_fshelp_find_file (path, &data->cmft, &fdiro, grub_ntfs_iterate_dir,
-			 0, GRUB_FSHELP_DIR);
+			 grub_ntfs_read_symlink, GRUB_FSHELP_DIR);
 
   if (grub_errno)
     goto fail;
 
-  grub_ntfs_iterate_dir (fdiro, iterate);
+  grub_ntfs_iterate_dir (fdiro, grub_ntfs_dir_iter, &ctx);
 
 fail:
   if ((fdiro) && (fdiro != &data->cmft))
@@ -931,7 +1052,7 @@ grub_ntfs_open (grub_file_t file, const char *name)
     goto fail;
 
   grub_fshelp_find_file (name, &data->cmft, &mft, grub_ntfs_iterate_dir,
-			 0, GRUB_FSHELP_REG);
+			 grub_ntfs_read_symlink, GRUB_FSHELP_REG);
 
   if (grub_errno)
     goto fail;
@@ -976,8 +1097,9 @@ grub_ntfs_read (grub_file_t file, char *buf, grub_size_t len)
   if (file->read_hook)
     mft->attr.save_pos = 1;
 
-  read_attr (&mft->attr, buf, file->offset, len, 1, file->read_hook);
-  return (grub_errno) ? 0 : len;
+  read_attr (&mft->attr, (grub_uint8_t *) buf, file->offset, len, 1,
+	     file->read_hook, file->read_hook_data);
+  return (grub_errno) ? -1 : (grub_ssize_t) len;
 }
 
 static grub_err_t
@@ -1004,7 +1126,7 @@ grub_ntfs_label (grub_device_t device, char **label)
 {
   struct grub_ntfs_data *data = 0;
   struct grub_fshelp_node *mft = 0;
-  char *pa;
+  grub_uint8_t *pa;
 
   grub_dl_ref (my_mod);
 
@@ -1022,7 +1144,7 @@ grub_ntfs_label (grub_device_t device, char **label)
 
   if (!mft->inode_read)
     {
-      mft->buf = grub_malloc (mft->data->mft_size << BLK_SHR);
+      mft->buf = grub_malloc (mft->data->mft_size << GRUB_NTFS_BLK_SHR);
       if (mft->buf == NULL)
 	goto fail;
 
@@ -1031,18 +1153,14 @@ grub_ntfs_label (grub_device_t device, char **label)
     }
 
   init_attr (&mft->attr, mft);
-  pa = find_attr (&mft->attr, AT_VOLUME_NAME);
+  pa = find_attr (&mft->attr, GRUB_NTFS_AT_VOLUME_NAME);
   if ((pa) && (pa[8] == 0) && (u32at (pa, 0x10)))
     {
-      char *buf;
       int len;
 
       len = u32at (pa, 0x10) / 2;
-      buf = grub_malloc (len * 4 + 1);
       pa += u16at (pa, 0x14);
-      *grub_utf16_to_utf8 ((grub_uint8_t *) buf, (grub_uint16_t *) pa, len) =
-	'\0';
-      *label = buf;
+      *label = get_utf8 (pa, len);
     }
 
 fail:
@@ -1079,14 +1197,15 @@ grub_ntfs_uuid (grub_device_t device, char **uuid)
       if (*uuid)
 	for (ptr = *uuid; *ptr; ptr++)
 	  *ptr = grub_toupper (*ptr);
+      free_file (&data->mmft);
+      free_file (&data->cmft);
+      grub_free (data);
     }
   else
     *uuid = NULL;
 
   grub_dl_unref (my_mod);
 
-  grub_free (data);
-
   return grub_errno;
 }
 
@@ -1101,6 +1220,7 @@ static struct grub_fs grub_ntfs_fs =
     .uuid = grub_ntfs_uuid,
 #ifdef GRUB_UTIL
     .reserved_first_sector = 1,
+    .blocklist_install = 1,
 #endif
     .next = 0
 };
