@@ -25,8 +25,39 @@
 #include <grub/env.h>
 #include <grub/i18n.h>
 
+union printf_arg
+{
+  /* Yes, type is also part of union as the moment we fill the value
+     we don't need to store its type anymore (when we'll need it, we'll
+     have format spec again. So save some space.  */
+  enum
+    {
+      INT, LONG, LONGLONG,
+      UNSIGNED_INT = 3, UNSIGNED_LONG, UNSIGNED_LONGLONG
+    } type;
+  long long ll;
+};
+
+struct printf_args
+{
+  union printf_arg prealloc[32];
+  union printf_arg *ptr;
+  grub_size_t count;
+};
+
+static void
+parse_printf_args (const char *fmt0, struct printf_args *args,
+		   va_list args_in);
 static int
-grub_vsnprintf_real (char *str, grub_size_t n, const char *fmt, va_list args);
+grub_vsnprintf_real (char *str, grub_size_t max_len, const char *fmt0,
+		     struct printf_args *args);
+
+static void
+free_printf_args (struct printf_args *args)
+{
+  if (args->ptr != args->prealloc)
+    grub_free (args->ptr);
+}
 
 static int
 grub_iswordseparator (int c)
@@ -64,23 +95,6 @@ grub_memmove (void *dest, const void *src, grub_size_t n)
   return dest;
 }
 
-#ifndef APPLE_CC
-void *memmove (void *dest, const void *src, grub_size_t n)
-  __attribute__ ((alias ("grub_memmove")));
-/* GCC emits references to memcpy() for struct copies etc.  */
-void *memcpy (void *dest, const void *src, grub_size_t n)
-  __attribute__ ((alias ("grub_memmove")));
-#else
-void *memcpy (void *dest, const void *src, grub_size_t n)
-{
-	return grub_memmove (dest, src, n);
-}
-void *memmove (void *dest, const void *src, grub_size_t n)
-{
-	return grub_memmove (dest, src, n);
-}
-#endif
-
 char *
 grub_strcpy (char *dest, const char *src)
 {
@@ -92,30 +106,6 @@ grub_strcpy (char *dest, const char *src)
   return dest;
 }
 
-char *
-grub_strncpy (char *dest, const char *src, int c)
-{
-  char *p = dest;
-
-  while ((*p++ = *src++) != '\0' && --c)
-    ;
-
-  return dest;
-}
-
-char *
-grub_stpcpy (char *dest, const char *src)
-{
-  char *d = dest;
-  const char *s = src;
-
-  do
-    *d++ = *s;
-  while (*s++ != '\0');
-
-  return d - 1;
-}
-
 int
 grub_printf (const char *fmt, ...)
 {
@@ -148,7 +138,7 @@ grub_puts_ (const char *s)
   return grub_puts (_(s));
 }
 
-#if defined (APPLE_CC) && ! defined (GRUB_UTIL)
+#if defined (__APPLE__) && ! defined (GRUB_UTIL)
 int
 grub_err_printf (const char *fmt, ...)
 {
@@ -163,7 +153,7 @@ grub_err_printf (const char *fmt, ...)
 }
 #endif
 
-#if ! defined (APPLE_CC) && ! defined (GRUB_UTIL)
+#if ! defined (__APPLE__) && ! defined (GRUB_UTIL)
 int grub_err_printf (const char *fmt, ...)
 __attribute__ ((alias("grub_printf")));
 #endif
@@ -191,15 +181,16 @@ grub_real_dprintf (const char *file, const int line, const char *condition,
 #define PREALLOC_SIZE 255
 
 int
-grub_vprintf (const char *fmt, va_list args)
+grub_vprintf (const char *fmt, va_list ap)
 {
   grub_size_t s;
   static char buf[PREALLOC_SIZE + 1];
   char *curbuf = buf;
-  va_list ap2;
-  va_copy (ap2, args);
+  struct printf_args args;
 
-  s = grub_vsnprintf_real (buf, PREALLOC_SIZE, fmt, args);
+  parse_printf_args (fmt, &args, ap);
+
+  s = grub_vsnprintf_real (buf, PREALLOC_SIZE, fmt, &args);
   if (s > PREALLOC_SIZE)
     {
       curbuf = grub_malloc (s + 1);
@@ -210,24 +201,27 @@ grub_vprintf (const char *fmt, va_list args)
 	  buf[PREALLOC_SIZE - 2] = '.';
 	  buf[PREALLOC_SIZE - 1] = '.';
 	  buf[PREALLOC_SIZE] = 0;
+	  curbuf = buf;
 	}
       else
-	s = grub_vsnprintf_real (curbuf, s, fmt, ap2);
+	s = grub_vsnprintf_real (curbuf, s, fmt, &args);
     }
 
+  free_printf_args (&args);
+
   grub_xputs (curbuf);
 
   if (curbuf != buf)
     grub_free (curbuf);
-  
+
   return s;
 }
 
 int
 grub_memcmp (const void *s1, const void *s2, grub_size_t n)
 {
-  const char *t1 = s1;
-  const char *t2 = s2;
+  const grub_uint8_t *t1 = s1;
+  const grub_uint8_t *t2 = s2;
 
   while (n--)
     {
@@ -240,15 +234,6 @@ grub_memcmp (const void *s1, const void *s2, grub_size_t n)
 
   return 0;
 }
-#ifndef APPLE_CC
-int memcmp (const void *s1, const void *s2, grub_size_t n)
-  __attribute__ ((alias ("grub_memcmp")));
-#else
-int memcmp (const void *s1, const void *s2, grub_size_t n)
-{
-  return grub_memcmp (s1, s2, n);
-}
-#endif
 
 int
 grub_strcmp (const char *s1, const char *s2)
@@ -262,7 +247,7 @@ grub_strcmp (const char *s1, const char *s2)
       s2++;
     }
 
-  return (int) *s1 - (int) *s2;
+  return (int) (grub_uint8_t) *s1 - (int) (grub_uint8_t) *s2;
 }
 
 int
@@ -280,7 +265,7 @@ grub_strncmp (const char *s1, const char *s2, grub_size_t n)
       s2++;
     }
 
-  return (int) *s1 - (int) *s2;
+  return (int) (grub_uint8_t) *s1 - (int) (grub_uint8_t)  *s2;
 }
 
 char *
@@ -311,52 +296,6 @@ grub_strrchr (const char *s, int c)
   return p;
 }
 
-/* Copied from gnulib.
-   Written by Bruno Haible <bruno@clisp.org>, 2005. */
-char *
-grub_strstr (const char *haystack, const char *needle)
-{
-  /* Be careful not to look at the entire extent of haystack or needle
-     until needed.  This is useful because of these two cases:
-       - haystack may be very long, and a match of needle found early,
-       - needle may be very long, and not even a short initial segment of
-       needle may be found in haystack.  */
-  if (*needle != '\0')
-    {
-      /* Speed up the following searches of needle by caching its first
-	 character.  */
-      char b = *needle++;
-
-      for (;; haystack++)
-	{
-	  if (*haystack == '\0')
-	    /* No match.  */
-	    return NULL;
-	  if (*haystack == b)
-	    /* The first character matches.  */
-	    {
-	      const char *rhaystack = haystack + 1;
-	      const char *rneedle = needle;
-
-	      for (;; rhaystack++, rneedle++)
-		{
-		  if (*rneedle == '\0')
-		    /* Found a match.  */
-		    return (char *) haystack;
-		  if (*rhaystack == '\0')
-		    /* No match.  */
-		    return NULL;
-		  if (*rhaystack != *rneedle)
-		    /* Nothing in this round.  */
-		    break;
-		}
-	    }
-	}
-    }
-  else
-    return (char *) haystack;
-}
-
 int
 grub_strword (const char *haystack, const char *needle)
 {
@@ -400,24 +339,19 @@ grub_isspace (int c)
   return (c == '\n' || c == '\r' || c == ' ' || c == '\t');
 }
 
-int
-grub_isprint (int c)
-{
-  return (c >= ' ' && c <= '~');
-}
-
-
 unsigned long
 grub_strtoul (const char *str, char **end, int base)
 {
   unsigned long long num;
 
   num = grub_strtoull (str, end, base);
+#if GRUB_CPU_SIZEOF_LONG != 8
   if (num > ~0UL)
     {
-      grub_error (GRUB_ERR_OUT_OF_RANGE, "overflow is detected");
+      grub_error (GRUB_ERR_OUT_OF_RANGE, N_("overflow is detected"));
       return ~0UL;
     }
+#endif
 
   return (unsigned long) num;
 }
@@ -429,7 +363,8 @@ grub_strtoull (const char *str, char **end, int base)
   int found = 0;
 
   /* Skip white spaces.  */
-  while (*str && grub_isspace (*str))
+  /* grub_isspace checks that *str != '\0'.  */
+  while (grub_isspace (*str))
     str++;
 
   /* Guess the base, if not specified. The prefix `0x' means 16, and
@@ -468,7 +403,8 @@ grub_strtoull (const char *str, char **end, int base)
       /* NUM * BASE + DIGIT > ~0ULL */
       if (num > grub_divmod64 (~0ULL - digit, base, 0))
 	{
-	  grub_error (GRUB_ERR_OUT_OF_RANGE, "overflow is detected");
+	  grub_error (GRUB_ERR_OUT_OF_RANGE,
+		      N_("overflow is detected"));
 	  return ~0ULL;
 	}
 
@@ -478,7 +414,8 @@ grub_strtoull (const char *str, char **end, int base)
 
   if (! found)
     {
-      grub_error (GRUB_ERR_BAD_NUMBER, "unrecognized number");
+      grub_error (GRUB_ERR_BAD_NUMBER,
+		  N_("unrecognized number"));
       return 0;
     }
 
@@ -520,6 +457,15 @@ grub_strndup (const char *s, grub_size_t n)
   return p;
 }
 
+/* clang detects that we're implementing here a memset so it decides to
+   optimise and calls memset resulting in infinite recursion. With volatile
+   we make it not optimise in this way.  */
+#ifdef __clang__
+#define VOLATILE_CLANG volatile
+#else
+#define VOLATILE_CLANG
+#endif
+
 void *
 grub_memset (void *s, int c, grub_size_t len)
 {
@@ -536,13 +482,13 @@ grub_memset (void *s, int c, grub_size_t len)
 
       while (len > 0 && (((grub_addr_t) p) & (sizeof (unsigned long) - 1)))
 	{
-	  *(grub_uint8_t *) p = pattern8;
+	  *(VOLATILE_CLANG grub_uint8_t *) p = pattern8;
 	  p = (grub_uint8_t *) p + 1;
 	  len--;
 	}
       while (len >= sizeof (unsigned long))
 	{
-	  *(unsigned long *) p = patternl;
+	  *(VOLATILE_CLANG unsigned long *) p = patternl;
 	  p = (unsigned long *) p + 1;
 	  len -= sizeof (unsigned long);
 	}
@@ -550,22 +496,13 @@ grub_memset (void *s, int c, grub_size_t len)
 
   while (len > 0)
     {
-      *(grub_uint8_t *) p = pattern8;
+      *(VOLATILE_CLANG grub_uint8_t *) p = pattern8;
       p = (grub_uint8_t *) p + 1;
       len--;
     }
 
   return s;
 }
-#ifndef APPLE_CC
-void *memset (void *s, int c, grub_size_t n)
-  __attribute__ ((alias ("grub_memset")));
-#else
-void *memset (void *s, int c, grub_size_t n)
-{
-  return grub_memset (s, c, n);
-}
-#endif
 
 grub_size_t
 grub_strlen (const char *s)
@@ -597,7 +534,7 @@ grub_reverse (char *str)
 
 /* Divide N by D, return the quotient, and store the remainder in *R.  */
 grub_uint64_t
-grub_divmod64_full (grub_uint64_t n, grub_uint64_t d, grub_uint64_t *r)
+grub_divmod64 (grub_uint64_t n, grub_uint64_t d, grub_uint64_t *r)
 {
   /* This algorithm is typically implemented by hardware. The idea
      is to get the highest bit in N, 64 times, by keeping
@@ -607,6 +544,11 @@ grub_divmod64_full (grub_uint64_t n, grub_uint64_t d, grub_uint64_t *r)
   grub_uint64_t q = 0;
   grub_uint64_t m = 0;
 
+  /* ARM and IA64 don't have a fast 32-bit division.
+     Using that code would just make us use software division routines, calling
+     ourselves indirectly and hence getting infinite recursion.
+  */
+#if !GRUB_DIVISION_IN_SOFTWARE
   /* Skip the slow computation if 32-bit arithmetic is possible.  */
   if (n < 0xffffffff && d < 0xffffffff)
     {
@@ -615,6 +557,7 @@ grub_divmod64_full (grub_uint64_t n, grub_uint64_t d, grub_uint64_t *r)
 
       return ((grub_uint32_t) n) / (grub_uint32_t) d;
     }
+#endif
 
   while (bits--)
     {
@@ -641,7 +584,7 @@ grub_divmod64_full (grub_uint64_t n, grub_uint64_t d, grub_uint64_t *r)
 
 /* Convert a long long value to a string. This function avoids 64-bit
    modular arithmetic or divisions.  */
-static char *
+static inline char *
 grub_lltoa (char *str, int c, unsigned long long n)
 {
   unsigned base = (c == 'x') ? 16 : 10;
@@ -666,7 +609,7 @@ grub_lltoa (char *str, int c, unsigned long long n)
     /* BASE == 10 */
     do
       {
-	unsigned m;
+	grub_uint64_t m;
 
 	n = grub_divmod64 (n, 10, &m);
 	*p++ = m + '0';
@@ -679,237 +622,384 @@ grub_lltoa (char *str, int c, unsigned long long n)
   return p;
 }
 
-static int
-grub_vsnprintf_real (char *str, grub_size_t max_len, const char *fmt, va_list args)
+static void
+parse_printf_args (const char *fmt0, struct printf_args *args,
+		   va_list args_in)
 {
+  const char *fmt;
   char c;
-  grub_size_t count = 0;
-  auto void write_char (unsigned char ch);
-  auto void write_str (const char *s);
-  auto void write_fill (const char ch, int n);
+  grub_size_t n = 0;
 
-  void write_char (unsigned char ch)
-    {
-      if (count < max_len)
-	*str++ = ch;
+  args->count = 0;
 
-      count++;
-    }
-
-  void write_str (const char *s)
-    {
-      while (*s)
-	write_char (*s++);
-    }
-
-  void write_fill (const char ch, int n)
-    {
-      int i;
-      for (i = 0; i < n; i++)
-	write_char (ch);
-    }
+  COMPILE_TIME_ASSERT (sizeof (int) == sizeof (grub_uint32_t));
+  COMPILE_TIME_ASSERT (sizeof (int) <= sizeof (long long));
+  COMPILE_TIME_ASSERT (sizeof (long) <= sizeof (long long));
+  COMPILE_TIME_ASSERT (sizeof (long long) == sizeof (void *)
+		       || sizeof (int) == sizeof (void *));
 
+  fmt = fmt0;
   while ((c = *fmt++) != 0)
     {
       if (c != '%')
-	write_char (c);
-      else
+	continue;
+
+      if (*fmt =='-')
+	fmt++;
+
+      while (grub_isdigit (*fmt))
+	fmt++;
+
+      if (*fmt == '$')
+	fmt++;
+
+      if (*fmt =='-')
+	fmt++;
+
+      while (grub_isdigit (*fmt))
+	fmt++;
+
+      if (*fmt =='.')
+	fmt++;
+
+      while (grub_isdigit (*fmt))
+	fmt++;
+
+      c = *fmt++;
+      if (c == 'l')
+	c = *fmt++;
+      if (c == 'l')
+	c = *fmt++;
+
+      switch (c)
 	{
-	  char tmp[32];
-	  char *p;
-	  unsigned int format1 = 0;
-	  unsigned int format2 = ~ 0U;
-	  char zerofill = ' ';
-	  int rightfill = 0;
-	  int n;
-	  int longfmt = 0;
-	  int longlongfmt = 0;
-	  int unsig = 0;
-
-	  if (*fmt && *fmt =='-')
-	    {
-	      rightfill = 1;
-	      fmt++;
-	    }
-
-	  p = (char *) fmt;
-	  /* Read formatting parameters.  */
-	  while (*p && grub_isdigit (*p))
-	    p++;
-
-	  if (p > fmt)
-	    {
-	      char s[p - fmt + 1];
-	      grub_strncpy (s, fmt, p - fmt);
-	      s[p - fmt] = 0;
-	      if (s[0] == '0')
-		zerofill = '0';
-	      format1 = grub_strtoul (s, 0, 10);
-	      fmt = p;
-	    }
-
-	  if (*p && *p == '.')
-	    {
-	      p++;
-	      fmt++;
-	      while (*p && grub_isdigit (*p))
-		p++;
-
-	      if (p > fmt)
-		{
-		  char fstr[p - fmt + 1];
-		  grub_strncpy (fstr, fmt, p - fmt);
-		  fstr[p - fmt] = 0;
-		  format2 = grub_strtoul (fstr, 0, 10);
-		  fmt = p;
-		}
-	    }
-
-	  c = *fmt++;
-	  if (c == 'l')
-	    {
-	      longfmt = 1;
-	      c = *fmt++;
-	      if (c == 'l')
-		{
-		  longlongfmt = 1;
-		  c = *fmt++;
-		}
-	    }
-
-	  switch (c)
-	    {
-	    case 'p':
-	      write_str ("0x");
-	      c = 'x';
-	      longlongfmt |= (sizeof (void *) == sizeof (long long));
-	      /* Fall through. */
-	    case 'x':
-	    case 'u':
-	      unsig = 1;
-	      /* Fall through. */
-	    case 'd':
-	      if (longlongfmt)
-		{
-		  long long ll;
-
-		  ll = va_arg (args, long long);
-		  grub_lltoa (tmp, c, ll);
-		}
-	      else if (longfmt && unsig)
-		{
-		  unsigned long l = va_arg (args, unsigned long);
-		  grub_lltoa (tmp, c, l);
-		}
-	      else if (longfmt)
-		{
-		  long l = va_arg (args, long);
-		  grub_lltoa (tmp, c, l);
-		}
-	      else if (unsig)
-		{
-		  unsigned u = va_arg (args, unsigned);
-		  grub_lltoa (tmp, c, u);
-		}
-	      else
-		{
-		  n = va_arg (args, int);
-		  grub_lltoa (tmp, c, n);
-		}
-	      if (! rightfill && grub_strlen (tmp) < format1)
-		write_fill (zerofill, format1 - grub_strlen (tmp));
-	      write_str (tmp);
-	      if (rightfill && grub_strlen (tmp) < format1)
-		write_fill (zerofill, format1 - grub_strlen (tmp));
-	      break;
-
-	    case 'c':
-	      n = va_arg (args, int);
-	      write_char (n & 0xff);
-	      break;
-
-	    case 'C':
-	      {
-		grub_uint32_t code = va_arg (args, grub_uint32_t);
-		int shift;
-		unsigned mask;
-
-		if (code <= 0x7f)
-		  {
-		    shift = 0;
-		    mask = 0;
-		  }
-		else if (code <= 0x7ff)
-		  {
-		    shift = 6;
-		    mask = 0xc0;
-		  }
-		else if (code <= 0xffff)
-		  {
-		    shift = 12;
-		    mask = 0xe0;
-		  }
-		else if (code <= 0x1fffff)
-		  {
-		    shift = 18;
-		    mask = 0xf0;
-		  }
-		else if (code <= 0x3ffffff)
-		  {
-		    shift = 24;
-		    mask = 0xf8;
-		  }
-		else if (code <= 0x7fffffff)
-		  {
-		    shift = 30;
-		    mask = 0xfc;
-		  }
-		else
-		  {
-		    code = '?';
-		    shift = 0;
-		    mask = 0;
-		  }
-
-		write_char (mask | (code >> shift));
-
-		for (shift -= 6; shift >= 0; shift -= 6)
-		  write_char (0x80 | (0x3f & (code >> shift)));
-	      }
-	      break;
-
-	    case 's':
-	      p = va_arg (args, char *);
-	      if (p)
-		{
-		  grub_size_t len = 0;
-		  while (len < format2 && p[len])
-		    len++;
-
-		  if (!rightfill && len < format1)
-		    write_fill (zerofill, format1 - len);
-
-		  grub_size_t i;
-		  for (i = 0; i < len; i++)
-		    write_char (*p++);
-
-		  if (rightfill && len < format1)
-		    write_fill (zerofill, format1 - len);
-		}
-	      else
-		write_str ("(null)");
-
-	      break;
-
-	    default:
-	      write_char (c);
-	      break;
-	    }
+	case 'p':
+	case 'x':
+	case 'u':
+	case 'd':
+	case 'c':
+	case 'C':
+	case 's':
+	  args->count++;
+	  break;
 	}
     }
 
-  *str = '\0';
+  if (args->count <= ARRAY_SIZE (args->prealloc))
+    args->ptr = args->prealloc;
+  else
+    {
+      args->ptr = grub_malloc (args->count * sizeof (args->ptr[0]));
+      if (!args->ptr)
+	{
+	  grub_errno = GRUB_ERR_NONE;
+	  args->ptr = args->prealloc;
+	  args->count = ARRAY_SIZE (args->prealloc);
+	}
+    }
 
+  grub_memset (args->ptr, 0, args->count * sizeof (args->ptr[0]));
+
+  fmt = fmt0;
+  n = 0;
+  while ((c = *fmt++) != 0)
+    {
+      int longfmt = 0;
+      grub_size_t curn;
+      const char *p;
+
+      if (c != '%')
+	continue;
+
+      curn = n++;
+
+      if (*fmt =='-')
+	fmt++;
+
+      p = fmt;
+
+      while (grub_isdigit (*fmt))
+	fmt++;
+
+      if (*fmt == '$')
+	{
+	  curn = grub_strtoull (p, 0, 10) - 1;
+	  fmt++;
+	}
+
+      if (*fmt =='-')
+	fmt++;
+
+      while (grub_isdigit (*fmt))
+	fmt++;
+
+      if (*fmt =='.')
+	fmt++;
+
+      while (grub_isdigit (*fmt))
+	fmt++;
+
+      c = *fmt++;
+      if (c == '%')
+	{
+	  n--;
+	  continue;
+	}
+
+      if (c == 'l')
+	{
+	  c = *fmt++;
+	  longfmt = 1;
+	}
+      if (c == 'l')
+	{
+	  c = *fmt++;
+	  longfmt = 2;
+	}
+      if (curn >= args->count)
+	continue;
+      switch (c)
+	{
+	case 'x':
+	case 'u':
+	  args->ptr[curn].type = UNSIGNED_INT + longfmt;
+	  break;
+	case 'd':
+	  args->ptr[curn].type = INT + longfmt;
+	  break;
+	case 'p':
+	case 's':
+	  if (sizeof (void *) == sizeof (long long))
+	    args->ptr[curn].type = UNSIGNED_LONGLONG;
+	  else
+	    args->ptr[curn].type = UNSIGNED_INT;
+	  break;
+	case 'C':
+	case 'c':
+	  args->ptr[curn].type = INT;
+	  break;
+	}
+    }
+
+  for (n = 0; n < args->count; n++)
+    switch (args->ptr[n].type)
+      {
+      case INT:
+	args->ptr[n].ll = va_arg (args_in, int);
+	break;
+      case LONG:
+	args->ptr[n].ll = va_arg (args_in, long);
+	break;
+      case UNSIGNED_INT:
+	args->ptr[n].ll = va_arg (args_in, unsigned int);
+	break;
+      case UNSIGNED_LONG:
+	args->ptr[n].ll = va_arg (args_in, unsigned long);
+	break;
+      case LONGLONG:
+      case UNSIGNED_LONGLONG:
+	args->ptr[n].ll = va_arg (args_in, long long);
+	break;
+      }
+}
+
+static inline void __attribute__ ((always_inline))
+write_char (char *str, grub_size_t *count, grub_size_t max_len, unsigned char ch)
+{
+  if (*count < max_len)
+    str[*count] = ch;
+
+  (*count)++;
+}
+
+static int
+grub_vsnprintf_real (char *str, grub_size_t max_len, const char *fmt0,
+		     struct printf_args *args)
+{
+  char c;
+  grub_size_t n = 0;
+  grub_size_t count = 0;
+  const char *fmt;
+
+  fmt = fmt0;
+
+  while ((c = *fmt++) != 0)
+    {
+      unsigned int format1 = 0;
+      unsigned int format2 = ~ 0U;
+      char zerofill = ' ';
+      char rightfill = 0;
+      grub_size_t curn;
+
+      if (c != '%')
+	{
+	  write_char (str, &count, max_len,c);
+	  continue;
+	}
+
+      curn = n++;
+
+    rescan:;
+
+      if (*fmt =='-')
+	{
+	  rightfill = 1;
+	  fmt++;
+	}
+
+      /* Read formatting parameters.  */
+      if (grub_isdigit (*fmt))
+	{
+	  if (fmt[0] == '0')
+	    zerofill = '0';
+	  format1 = grub_strtoul (fmt, (char **) &fmt, 10);
+	}
+
+      if (*fmt == '.')
+	fmt++;
+
+      if (grub_isdigit (*fmt))
+	format2 = grub_strtoul (fmt, (char **) &fmt, 10);
+
+      if (*fmt == '$')
+	{
+	  curn = format1 - 1;
+	  fmt++;
+	  format1 = 0;
+	  format2 = ~ 0U;
+	  zerofill = ' ';
+	  rightfill = 0;
+
+	  goto rescan;
+	}
+
+      c = *fmt++;
+      if (c == 'l')
+	c = *fmt++;
+      if (c == 'l')
+	c = *fmt++;
+
+      if (c == '%')
+	{
+	  write_char (str, &count, max_len,c);
+	  n--;
+	  continue;
+	}
+
+      if (curn >= args->count)
+	continue;
+
+      long long curarg = args->ptr[curn].ll;
+
+      switch (c)
+	{
+	case 'p':
+	  write_char (str, &count, max_len, '0');
+	  write_char (str, &count, max_len, 'x');
+	  c = 'x';
+	  /* Fall through. */
+	case 'x':
+	case 'u':
+	case 'd':
+	  {
+	    char tmp[32];
+	    const char *p = tmp;
+	    grub_size_t len;
+	    grub_size_t fill;
+
+	    len = grub_lltoa (tmp, c, curarg) - tmp;
+	    fill = len < format1 ? format1 - len : 0;
+	    if (! rightfill)
+	      while (fill--)
+		write_char (str, &count, max_len, zerofill);
+	    while (*p)
+	      write_char (str, &count, max_len, *p++);
+	    if (rightfill)
+	      while (fill--)
+		write_char (str, &count, max_len, zerofill);
+	  }
+	  break;
+
+	case 'c':
+	  write_char (str, &count, max_len,curarg & 0xff);
+	  break;
+
+	case 'C':
+	  {
+	    grub_uint32_t code = curarg;
+	    int shift;
+	    unsigned mask;
+
+	    if (code <= 0x7f)
+	      {
+		shift = 0;
+		mask = 0;
+	      }
+	    else if (code <= 0x7ff)
+	      {
+		shift = 6;
+		mask = 0xc0;
+	      }
+	    else if (code <= 0xffff)
+	      {
+		shift = 12;
+		mask = 0xe0;
+	      }
+	    else if (code <= 0x10ffff)
+	      {
+		shift = 18;
+		mask = 0xf0;
+	      }
+	    else
+	      {
+		code = '?';
+		shift = 0;
+		mask = 0;
+	      }
+
+	    write_char (str, &count, max_len,mask | (code >> shift));
+
+	    for (shift -= 6; shift >= 0; shift -= 6)
+	      write_char (str, &count, max_len,0x80 | (0x3f & (code >> shift)));
+	  }
+	  break;
+
+	case 's':
+	  {
+	    grub_size_t len = 0;
+	    grub_size_t fill;
+	    const char *p = ((char *) (grub_addr_t) curarg) ? : "(null)";
+	    grub_size_t i;
+
+	    while (len < format2 && p[len])
+	      len++;
+
+	    fill = len < format1 ? format1 - len : 0;
+
+	    if (!rightfill)
+	      while (fill--)
+		write_char (str, &count, max_len, zerofill);
+
+	    for (i = 0; i < len; i++)
+	      write_char (str, &count, max_len,*p++);
+
+	    if (rightfill)
+	      while (fill--)
+		write_char (str, &count, max_len, zerofill);
+	  }
+
+	  break;
+
+	default:
+	  write_char (str, &count, max_len,c);
+	  break;
+	}
+    }
+
+  if (count < max_len)
+    str[count] = '\0';
+  else
+    str[max_len] = '\0';
   return count;
 }
 
@@ -917,13 +1007,18 @@ int
 grub_vsnprintf (char *str, grub_size_t n, const char *fmt, va_list ap)
 {
   grub_size_t ret;
+  struct printf_args args;
 
   if (!n)
     return 0;
 
   n--;
 
-  ret = grub_vsnprintf_real (str, n, fmt, ap);
+  parse_printf_args (fmt, &args, ap);
+
+  ret = grub_vsnprintf_real (str, n, fmt, &args);
+
+  free_printf_args (&args);
 
   return ret < n ? ret : n;
 }
@@ -946,18 +1041,26 @@ grub_xvasprintf (const char *fmt, va_list ap)
 {
   grub_size_t s, as = PREALLOC_SIZE;
   char *ret;
+  struct printf_args args;
+
+  parse_printf_args (fmt, &args, ap);
 
   while (1)
     {
-      va_list ap2;
-      va_copy (ap2, ap);
       ret = grub_malloc (as + 1);
       if (!ret)
-	return NULL;
+	{
+	  free_printf_args (&args);
+	  return NULL;
+	}
+
+      s = grub_vsnprintf_real (ret, as, fmt, &args);
 
-      s = grub_vsnprintf_real (ret, as, fmt, ap2);
       if (s <= as)
-	return ret;
+	{
+	  free_printf_args (&args);
+	  return ret;
+	}
 
       grub_free (ret);
       as = s;
@@ -978,7 +1081,7 @@ grub_xasprintf (const char *fmt, ...)
 }
 
 /* Abort GRUB. This function does not return.  */
-void
+static void __attribute__ ((noreturn))
 grub_abort (void)
 {
   grub_printf ("\nAborted.");
@@ -994,26 +1097,56 @@ grub_abort (void)
   grub_exit ();
 }
 
-#if ! defined (APPLE_CC) && !defined (GRUB_UTIL)
-/* GCC emits references to abort().  */
-void abort (void) __attribute__ ((alias ("grub_abort")));
-#endif
-
-#if NEED_ENABLE_EXECUTE_STACK && !defined(GRUB_UTIL) && !defined(GRUB_MACHINE_EMU)
-/* Some gcc versions generate a call to this function
-   in trampolines for nested functions.  */
-void __enable_execute_stack (void *addr __attribute__ ((unused)))
+void
+grub_fatal (const char *fmt, ...)
 {
-}
-#endif
+  va_list ap;
 
-#if NEED_REGISTER_FRAME_INFO && !defined(GRUB_UTIL)
-void __register_frame_info (void)
-{
+  va_start (ap, fmt);
+  grub_vprintf (_(fmt), ap);
+  va_end (ap);
+
+  grub_refresh ();
+
+  grub_abort ();
 }
 
-void __deregister_frame_info (void)
+#if BOOT_TIME_STATS
+
+#include <grub/time.h>
+
+struct grub_boot_time *grub_boot_time_head;
+static struct grub_boot_time **boot_time_last = &grub_boot_time_head;
+
+void
+grub_real_boot_time (const char *file,
+		     const int line,
+		     const char *fmt, ...)
 {
+  struct grub_boot_time *n;
+  va_list args;
+
+  grub_error_push ();
+  n = grub_malloc (sizeof (*n));
+  if (!n)
+    {
+      grub_errno = 0;
+      grub_error_pop ();
+      return;
+    }
+  n->file = file;
+  n->line = line;
+  n->tp = grub_get_time_ms ();
+  n->next = 0;
+
+  va_start (args, fmt);
+  n->msg = grub_xvasprintf (fmt, args);    
+  va_end (args);
+
+  *boot_time_last = n;
+  boot_time_last = &n->next;
+
+  grub_errno = 0;
+  grub_error_pop ();
 }
 #endif
-
