@@ -27,19 +27,26 @@
 #include <grub/mips/relocator.h>
 #include <grub/memory.h>
 #include <grub/i18n.h>
+#include <grub/lib/cmdline.h>
+#include <grub/linux.h>
 
 GRUB_MOD_LICENSE ("GPLv3+");
 
+#pragma GCC diagnostic ignored "-Wcast-align"
+
 /* For frequencies.  */
-#include <grub/pci.h>
 #include <grub/machine/time.h>
 
-#ifdef GRUB_MACHINE_MIPS_YEELOONG
-/* This can be detected on runtime from PMON, but:
-     a) it wouldn't work when GRUB is the firmware
-   and
-     b) for now we only support Yeeloong anyway.  */
-#define LOONGSON_MACHTYPE "machtype=lemote-yeeloong-2f-8.9inches"
+#ifdef GRUB_MACHINE_MIPS_LOONGSON
+#include <grub/pci.h>
+#include <grub/machine/kernel.h>
+
+const char loongson_machtypes[][60] =
+  {
+    [GRUB_ARCH_MACHINE_YEELOONG] = "machtype=lemote-yeeloong-2f-8.9inches",
+    [GRUB_ARCH_MACHINE_FULOONG2F]  = "machtype=lemote-fuloong-2f-box",
+    [GRUB_ARCH_MACHINE_FULOONG2E]  = "machtype=lemote-fuloong-2e-unknown"
+  };
 #endif
 
 static grub_dl_t my_mod;
@@ -51,9 +58,16 @@ static grub_size_t linux_size;
 static struct grub_relocator *relocator;
 static grub_uint8_t *playground;
 static grub_addr_t target_addr, entry_addr;
+#ifdef GRUB_MACHINE_MIPS_QEMU_MIPS
+static char *params;
+#else
 static int linux_argc;
-static grub_off_t argv_off, envp_off;
+static grub_off_t argv_off;
+#ifdef GRUB_MACHINE_MIPS_LOONGSON
+static grub_off_t envp_off;
+#endif
 static grub_off_t rd_addr_arg_off, rd_size_arg_off;
+#endif
 static int initrd_loaded = 0;
 
 static grub_err_t
@@ -61,11 +75,43 @@ grub_linux_boot (void)
 {
   struct grub_relocator32_state state;
 
+  grub_memset (&state, 0, sizeof (state));
+
   /* Boot the kernel.  */
   state.gpr[1] = entry_addr;
+
+#ifdef GRUB_MACHINE_MIPS_QEMU_MIPS
+  {
+    grub_err_t err;
+    grub_relocator_chunk_t ch;
+    grub_uint32_t *memsize;
+    grub_uint32_t *magic;
+    char *str;
+
+    err = grub_relocator_alloc_chunk_addr (relocator, &ch,
+					   ((16 << 20) - 264),
+					   grub_strlen (params) + 1 + 8);
+    if (err)
+      return err;
+    memsize = get_virtual_current_address (ch);
+    magic = memsize + 1;
+    *memsize = grub_mmap_get_lower ();
+    *magic = 0x12345678;
+    str = (char *) (magic + 1);
+    grub_strcpy (str, params);
+  }
+#endif  
+
+#ifndef GRUB_MACHINE_MIPS_QEMU_MIPS
   state.gpr[4] = linux_argc;
   state.gpr[5] = target_addr + argv_off;
+#ifdef GRUB_MACHINE_MIPS_LOONGSON
   state.gpr[6] = target_addr + envp_off;
+#else
+  state.gpr[6] = 0;
+#endif
+  state.gpr[7] = 0;
+#endif
   state.jumpreg = 1;
   grub_relocator32_boot (relocator, state);
 
@@ -78,13 +124,19 @@ grub_linux_unload (void)
   grub_relocator_unload (relocator);
   grub_dl_unref (my_mod);
 
+#ifdef GRUB_MACHINE_MIPS_QEMU_MIPS
+  grub_free (params);
+  params = 0;
+#endif
+
   loaded = 0;
 
   return GRUB_ERR_NONE;
 }
 
 static grub_err_t
-grub_linux_load32 (grub_elf_t elf, void **extra_mem, grub_size_t extra_size)
+grub_linux_load32 (grub_elf_t elf, const char *filename,
+		   void **extra_mem, grub_size_t extra_size)
 {
   Elf32_Addr base;
   int extraoff;
@@ -120,26 +172,12 @@ grub_linux_load32 (grub_elf_t elf, void **extra_mem, grub_size_t extra_size)
   *extra_mem = playground + extraoff;
 
   /* Now load the segments into the area we claimed.  */
-  auto grub_err_t offset_phdr (Elf32_Phdr *phdr, grub_addr_t *addr, int *do_load);
-  grub_err_t offset_phdr (Elf32_Phdr *phdr, grub_addr_t *addr, int *do_load)
-    {
-      if (phdr->p_type != PT_LOAD)
-	{
-	  *do_load = 0;
-	  return 0;
-	}
-      *do_load = 1;
-
-      /* Linux's program headers incorrectly contain virtual addresses.
-       * Translate those to physical, and offset to the area we claimed.  */
-      *addr = (grub_addr_t) (phdr->p_paddr - base + playground);
-      return 0;
-    }
-  return grub_elf32_load (elf, offset_phdr, 0, 0);
+  return grub_elf32_load (elf, filename, playground - base, GRUB_ELF_LOAD_FLAGS_NONE, 0, 0);
 }
 
 static grub_err_t
-grub_linux_load64 (grub_elf_t elf, void **extra_mem, grub_size_t extra_size)
+grub_linux_load64 (grub_elf_t elf, const char *filename,
+		   void **extra_mem, grub_size_t extra_size)
 {
   Elf64_Addr base;
   int extraoff;
@@ -175,21 +213,7 @@ grub_linux_load64 (grub_elf_t elf, void **extra_mem, grub_size_t extra_size)
   *extra_mem = playground + extraoff;
 
   /* Now load the segments into the area we claimed.  */
-  auto grub_err_t offset_phdr (Elf64_Phdr *phdr, grub_addr_t *addr, int *do_load);
-  grub_err_t offset_phdr (Elf64_Phdr *phdr, grub_addr_t *addr, int *do_load)
-    {
-      if (phdr->p_type != PT_LOAD)
-	{
-	  *do_load = 0;
-	  return 0;
-	}
-      *do_load = 1;
-      /* Linux's program headers incorrectly contain virtual addresses.
-       * Translate those to physical, and offset to the area we claimed.  */
-      *addr = (grub_addr_t) (phdr->p_paddr - base + playground);
-      return 0;
-    }
-  return grub_elf64_load (elf, offset_phdr, 0, 0);
+  return grub_elf64_load (elf, filename, playground - base, GRUB_ELF_LOAD_FLAGS_NONE, 0, 0);
 }
 
 static grub_err_t
@@ -197,15 +221,21 @@ grub_cmd_linux (grub_command_t cmd __attribute__ ((unused)),
 		int argc, char *argv[])
 {
   grub_elf_t elf = 0;
-  int i;
   int size;
   void *extra = NULL;
-  grub_uint32_t *linux_argv, *linux_envp;
-  char *linux_args, *linux_envs;
+#ifndef GRUB_MACHINE_MIPS_QEMU_MIPS
+  int i;
+  grub_uint32_t *linux_argv;
+  char *linux_args;
+#endif
   grub_err_t err;
+#ifdef GRUB_MACHINE_MIPS_LOONGSON
+  char *linux_envs;
+  grub_uint32_t *linux_envp;
+#endif
 
   if (argc == 0)
-    return grub_error (GRUB_ERR_BAD_ARGUMENT, "no kernel specified");
+    return grub_error (GRUB_ERR_BAD_ARGUMENT, N_("filename expected"));
 
   elf = grub_elf_open (argv[0]);
   if (! elf)
@@ -215,16 +245,19 @@ grub_cmd_linux (grub_command_t cmd __attribute__ ((unused)),
     {
       grub_elf_close (elf);
       return grub_error (GRUB_ERR_UNKNOWN_OS,
-			 "this ELF file is not of the right type\n");
+			 N_("this ELF file is not of the right type"));
     }
 
   /* Release the previously used memory.  */
   grub_loader_unset ();
   loaded = 0;
 
+#ifdef GRUB_MACHINE_MIPS_QEMU_MIPS
+  size = 0;
+#else
   /* For arguments.  */
   linux_argc = argc;
-#ifdef LOONGSON_MACHTYPE
+#ifdef GRUB_MACHINE_MIPS_LOONGSON
   linux_argc++;
 #endif
   /* Main arguments.  */
@@ -239,8 +272,8 @@ grub_cmd_linux (grub_command_t cmd __attribute__ ((unused)),
   /* Normal arguments.  */
   for (i = 1; i < argc; i++)
     size += ALIGN_UP (grub_strlen (argv[i]) + 1, 4);
-#ifdef LOONGSON_MACHTYPE
-  size += ALIGN_UP (sizeof (LOONGSON_MACHTYPE), 4);
+#ifdef GRUB_MACHINE_MIPS_LOONGSON
+  size += ALIGN_UP (sizeof (loongson_machtypes[0]), 4);
 #endif
 
   /* rd arguments.  */
@@ -254,20 +287,35 @@ grub_cmd_linux (grub_command_t cmd __attribute__ ((unused)),
     + ALIGN_UP (sizeof ("highmemsize=XXXXXXXXXXXXXXXXXXXX"), 4)
     + ALIGN_UP (sizeof ("busclock=XXXXXXXXXX"), 4)
     + ALIGN_UP (sizeof ("cpuclock=XXXXXXXXXX"), 4);
+#endif
 
   if (grub_elf_is_elf32 (elf))
-    err = grub_linux_load32 (elf, &extra, size);
+    err = grub_linux_load32 (elf, argv[0], &extra, size);
   else
   if (grub_elf_is_elf64 (elf))
-    err = grub_linux_load64 (elf, &extra, size);
+    err = grub_linux_load64 (elf, argv[0], &extra, size);
   else
-    err = grub_error (GRUB_ERR_BAD_FILE_TYPE, "unknown ELF class");
+    err = grub_error (GRUB_ERR_BAD_OS, N_("invalid arch-dependent ELF magic"));
 
   grub_elf_close (elf);
 
   if (err)
     return err;
 
+#ifdef GRUB_MACHINE_MIPS_QEMU_MIPS
+  /* Create kernel command line.  */
+  size = grub_loader_cmdline_size(argc, argv);
+  params = grub_malloc (size + sizeof (LINUX_IMAGE));
+  if (! params)
+    {
+      grub_linux_unload ();
+      return grub_errno;
+    }
+
+  grub_memcpy (params, LINUX_IMAGE, sizeof (LINUX_IMAGE));
+  grub_create_loader_cmdline (argc, argv, params + sizeof (LINUX_IMAGE) - 1,
+			      size);
+#else
   linux_argv = extra;
   argv_off = (grub_uint8_t *) linux_argv - (grub_uint8_t *) playground;
   extra = linux_argv + (linux_argc + 1 + 2);
@@ -279,14 +327,20 @@ grub_cmd_linux (grub_command_t cmd __attribute__ ((unused)),
   linux_argv++;
   linux_args += ALIGN_UP (sizeof ("a0"), 4);
 
-#ifdef LOONGSON_MACHTYPE
-  /* In Loongson platform, it is the responsibility of the bootloader/firmware
-     to supply the OS kernel with machine type information.  */
-  grub_memcpy (linux_args, LOONGSON_MACHTYPE, sizeof (LOONGSON_MACHTYPE));
-  *linux_argv = (grub_uint8_t *) linux_args - (grub_uint8_t *) playground
-    + target_addr;
-  linux_argv++;
-  linux_args += ALIGN_UP (sizeof (LOONGSON_MACHTYPE), 4);
+#ifdef GRUB_MACHINE_MIPS_LOONGSON
+  {
+    unsigned mtype = grub_arch_machine;
+    if (mtype >= ARRAY_SIZE (loongson_machtypes))
+      mtype = 0;
+    /* In Loongson platform, it is the responsibility of the bootloader/firmware
+       to supply the OS kernel with machine type information.  */
+    grub_memcpy (linux_args, loongson_machtypes[mtype],
+		 sizeof (loongson_machtypes[mtype]));
+    *linux_argv = (grub_uint8_t *) linux_args - (grub_uint8_t *) playground
+      + target_addr;
+    linux_argv++;
+    linux_args += ALIGN_UP (sizeof (loongson_machtypes[mtype]), 4);
+  }
 #endif
 
   for (i = 1; i < argc; i++)
@@ -313,6 +367,7 @@ grub_cmd_linux (grub_command_t cmd __attribute__ ((unused)),
 
   extra = linux_args;
 
+#ifdef GRUB_MACHINE_MIPS_LOONGSON
   linux_envp = extra;
   envp_off = (grub_uint8_t *) linux_envp - (grub_uint8_t *) playground;
   linux_envs = (char *) (linux_envp + 5);
@@ -340,8 +395,9 @@ grub_cmd_linux (grub_command_t cmd __attribute__ ((unused)),
     + target_addr;
   linux_envs += ALIGN_UP (grub_strlen (linux_envs) + 1, 4);
 
-
   linux_envp[4] = 0;
+#endif
+#endif
 
   grub_loader_set (grub_linux_boot, grub_linux_unload, 1);
   initrd_loaded = 0;
@@ -355,27 +411,25 @@ static grub_err_t
 grub_cmd_initrd (grub_command_t cmd __attribute__ ((unused)),
 		 int argc, char *argv[])
 {
-  grub_file_t file = 0;
-  grub_ssize_t size;
+  grub_size_t size = 0;
   void *initrd_src;
   grub_addr_t initrd_dest;
   grub_err_t err;
+  struct grub_linux_initrd_context initrd_ctx = { 0, 0, 0 };
 
   if (argc == 0)
-    return grub_error (GRUB_ERR_BAD_ARGUMENT, "no initrd specified");
+    return grub_error (GRUB_ERR_BAD_ARGUMENT, N_("filename expected"));
 
   if (!loaded)
-    return grub_error (GRUB_ERR_BAD_ARGUMENT, "you need to load Linux first.");
+    return grub_error (GRUB_ERR_BAD_ARGUMENT, N_("you need to load the kernel first"));
 
   if (initrd_loaded)
-    return grub_error (GRUB_ERR_BAD_ARGUMENT, "only one initrd can be loaded.");
+    return grub_error (GRUB_ERR_BAD_ARGUMENT, "only one initrd command can be issued.");
 
-  grub_file_filter_disable_compression ();
-  file = grub_file_open (argv[0]);
-  if (! file)
-    return grub_errno;
+  if (grub_initrd_init (argc, argv, &initrd_ctx))
+    goto fail;
 
-  size = grub_file_size (file);
+  size = grub_get_initrd_size (&initrd_ctx);
 
   {
     grub_relocator_chunk_t ch;
@@ -385,25 +439,29 @@ grub_cmd_initrd (grub_command_t cmd __attribute__ ((unused)),
 					    + linux_size + 0x10000,
 					    (0x10000000 - size),
 					    size, 0x10000,
-					    GRUB_RELOCATOR_PREFERENCE_NONE);
+					    GRUB_RELOCATOR_PREFERENCE_NONE, 0);
 
     if (err)
-      {
-	grub_file_close (file);
-	return err;
-      }
+      goto fail;
     initrd_src = get_virtual_current_address (ch);
     initrd_dest = get_physical_target_address (ch) | 0x80000000;
   }
 
-  if (grub_file_read (file, initrd_src, size) != size)
-    {
-      grub_error (GRUB_ERR_FILE_READ_ERROR, "couldn't read file");
-      grub_file_close (file);
-
-      return grub_errno;
-    }
+  if (grub_initrd_load (&initrd_ctx, argv, initrd_src))
+    goto fail;
 
+#ifdef GRUB_MACHINE_MIPS_QEMU_MIPS
+  {
+    char *tmp;
+    tmp = grub_xasprintf ("%s rd_start=0x%" PRIxGRUB_ADDR
+			  " rd_size=0x%" PRIxGRUB_ADDR, params,
+			  initrd_dest, size);
+    if (!tmp)
+      goto fail;
+    grub_free (params);
+    params = tmp;
+  }
+#else
   grub_snprintf ((char *) playground + rd_addr_arg_off,
 		 sizeof ("rd_start=0xXXXXXXXXXXXXXXXX"), "rd_start=0x%llx",
 		(unsigned long long) initrd_dest);
@@ -417,12 +475,14 @@ grub_cmd_initrd (grub_command_t cmd __attribute__ ((unused)),
   ((grub_uint32_t *) (playground + argv_off))[linux_argc]
     = target_addr + rd_size_arg_off;
   linux_argc++;
+#endif
 
   initrd_loaded = 1;
 
-  grub_file_close (file);
+ fail:
+  grub_initrd_close (&initrd_ctx);
 
-  return GRUB_ERR_NONE;
+  return grub_errno;
 }
 
 static grub_command_t cmd_linux, cmd_initrd;
