@@ -28,6 +28,8 @@
 #include <grub/i18n.h>
 #include <grub/memory.h>
 #include <grub/lib/cmdline.h>
+#include <grub/cache.h>
+#include <grub/linux.h>
 
 GRUB_MOD_LICENSE ("GPLv3+");
 
@@ -42,6 +44,7 @@ static grub_addr_t initrd_addr;
 static grub_size_t initrd_size;
 
 static grub_addr_t linux_addr;
+static grub_addr_t linux_entry;
 static grub_size_t linux_size;
 
 static char *linux_args;
@@ -49,49 +52,82 @@ static char *linux_args;
 typedef void (*kernel_entry_t) (void *, unsigned long, int (void *),
 				unsigned long, unsigned long);
 
+/* Context for grub_linux_claimmap_iterate.  */
+struct grub_linux_claimmap_iterate_ctx
+{
+  grub_addr_t target;
+  grub_size_t size;
+  grub_size_t align;
+  grub_addr_t found_addr;
+};
+
+/* Helper for grub_linux_claimmap_iterate.  */
+static int
+alloc_mem (grub_uint64_t addr, grub_uint64_t len, grub_memory_type_t type,
+	   void *data)
+{
+  struct grub_linux_claimmap_iterate_ctx *ctx = data;
+
+  grub_uint64_t end = addr + len;
+  addr = ALIGN_UP (addr, ctx->align);
+  ctx->target = ALIGN_UP (ctx->target, ctx->align);
+
+  /* Target above the memory chunk.  */
+  if (type != GRUB_MEMORY_AVAILABLE || ctx->target > end)
+    return 0;
+
+  /* Target inside the memory chunk.  */
+  if (ctx->target >= addr && ctx->target < end &&
+      ctx->size <= end - ctx->target)
+    {
+      if (grub_claimmap (ctx->target, ctx->size) == GRUB_ERR_NONE)
+	{
+	  ctx->found_addr = ctx->target;
+	  return 1;
+	}
+      grub_print_error ();
+    }
+  /* Target below the memory chunk.  */
+  if (ctx->target < addr && addr + ctx->size <= end)
+    {
+      if (grub_claimmap (addr, ctx->size) == GRUB_ERR_NONE)
+	{
+	  ctx->found_addr = addr;
+	  return 1;
+	}
+      grub_print_error ();
+    }
+  return 0;
+}
+
 static grub_addr_t
 grub_linux_claimmap_iterate (grub_addr_t target, grub_size_t size,
 			     grub_size_t align)
 {
-  grub_addr_t found_addr = (grub_addr_t) -1;
+  struct grub_linux_claimmap_iterate_ctx ctx = {
+    .target = target,
+    .size = size,
+    .align = align,
+    .found_addr = (grub_addr_t) -1
+  };
 
-  auto int NESTED_FUNC_ATTR alloc_mem (grub_uint64_t addr, grub_uint64_t len,
-				       grub_memory_type_t type);
-  int NESTED_FUNC_ATTR alloc_mem (grub_uint64_t addr, grub_uint64_t len,
-				  grub_memory_type_t type)
-  {
-    grub_uint64_t end = addr + len;
-    addr = ALIGN_UP (addr, align);
-    target = ALIGN_UP (target, align);
+  if (grub_ieee1275_test_flag (GRUB_IEEE1275_FLAG_FORCE_CLAIM))
+    {
+      grub_uint64_t addr = target;
+      if (addr < GRUB_IEEE1275_STATIC_HEAP_START
+	  + GRUB_IEEE1275_STATIC_HEAP_LEN)
+	addr = GRUB_IEEE1275_STATIC_HEAP_START
+	  + GRUB_IEEE1275_STATIC_HEAP_LEN;
+      addr = ALIGN_UP (addr, align);
+      if (grub_claimmap (addr, size) == GRUB_ERR_NONE)
+	return addr;
+      return (grub_addr_t) -1;
+    }
+	
 
-    /* Target above the memory chunk.  */
-    if (type != GRUB_MEMORY_AVAILABLE || target > end)
-      return 0;
+  grub_machine_mmap_iterate (alloc_mem, &ctx);
 
-    /* Target inside the memory chunk.  */
-    if (target >= addr && target < end && size <= end - target)
-      {
-	if (grub_claimmap (target, size) == GRUB_ERR_NONE)
-	  {
-	    found_addr = target;
-	    return 1;
-	  }
-      }
-    /* Target below the memory chunk.  */
-    if (target < addr && addr + size <= end)
-      {
-	if (grub_claimmap (addr, size) == GRUB_ERR_NONE)
-	  {
-	    found_addr = addr;
-	    return 1;
-	  }
-      }
-    return 0;
-  }
-
-  grub_machine_mmap_iterate (alloc_mem);
-
-  return found_addr;
+  return ctx.found_addr;
 }
 
 static grub_err_t
@@ -100,18 +136,19 @@ grub_linux_boot (void)
   kernel_entry_t linuxmain;
   grub_ssize_t actual;
 
+  grub_arch_sync_caches ((void *) linux_addr, linux_size);
   /* Set the command line arguments.  */
   grub_ieee1275_set_property (grub_ieee1275_chosen, "bootargs", linux_args,
 			      grub_strlen (linux_args) + 1, &actual);
 
-  grub_dprintf ("loader", "Entry point: 0x%x\n", linux_addr);
+  grub_dprintf ("loader", "Entry point: 0x%x\n", linux_entry);
   grub_dprintf ("loader", "Initrd at: 0x%x, size 0x%x\n", initrd_addr,
 		initrd_size);
   grub_dprintf ("loader", "Boot arguments: %s\n", linux_args);
   grub_dprintf ("loader", "Jumping to Linux...\n");
 
   /* Boot the kernel.  */
-  linuxmain = (kernel_entry_t) linux_addr;
+  linuxmain = (kernel_entry_t) linux_entry;
   linuxmain ((void *) initrd_addr, initrd_size, grub_ieee1275_entry_fn, 0, 0);
 
   return GRUB_ERR_NONE;
@@ -149,7 +186,7 @@ grub_linux_unload (void)
 }
 
 static grub_err_t
-grub_linux_load32 (grub_elf_t elf)
+grub_linux_load32 (grub_elf_t elf, const char *filename)
 {
   Elf32_Addr base_addr;
   grub_addr_t seg_addr;
@@ -177,27 +214,15 @@ grub_linux_load32 (grub_elf_t elf)
   if (seg_addr == (grub_addr_t) -1)
     return grub_error (GRUB_ERR_OUT_OF_MEMORY, "couldn't claim memory");
 
-  linux_addr = seg_addr + offset;
+  linux_entry = seg_addr + offset;
+  linux_addr = seg_addr;
 
   /* Now load the segments into the area we claimed.  */
-  auto grub_err_t offset_phdr (Elf32_Phdr *phdr, grub_addr_t *addr, int *do_load);
-  grub_err_t offset_phdr (Elf32_Phdr *phdr, grub_addr_t *addr, int *do_load)
-    {
-      if (phdr->p_type != PT_LOAD)
-	{
-	  *do_load = 0;
-	  return 0;
-	}
-      *do_load = 1;
-
-      *addr = (phdr->p_paddr - base_addr) + seg_addr;
-      return 0;
-    }
-  return grub_elf32_load (elf, offset_phdr, 0, 0);
+  return grub_elf32_load (elf, filename, (void *) (seg_addr - base_addr), GRUB_ELF_LOAD_FLAGS_30BITS, 0, 0);
 }
 
 static grub_err_t
-grub_linux_load64 (grub_elf_t elf)
+grub_linux_load64 (grub_elf_t elf, const char *filename)
 {
   Elf64_Addr base_addr;
   grub_addr_t seg_addr;
@@ -223,23 +248,11 @@ grub_linux_load64 (grub_elf_t elf)
   if (seg_addr == (grub_addr_t) -1)
     return grub_error (GRUB_ERR_OUT_OF_MEMORY, "couldn't claim memory");
 
-  linux_addr = seg_addr + offset;
+  linux_entry = seg_addr + offset;
+  linux_addr = seg_addr;
 
   /* Now load the segments into the area we claimed.  */
-  auto grub_err_t offset_phdr (Elf64_Phdr *phdr, grub_addr_t *addr, int *do_load);
-  grub_err_t offset_phdr (Elf64_Phdr *phdr, grub_addr_t *addr, int *do_load)
-    {
-      if (phdr->p_type != PT_LOAD)
-	{
-	  *do_load = 0;
-	  return 0;
-	}
-      *do_load = 1;
-
-      *addr = (phdr->p_paddr - base_addr) + seg_addr;
-      return 0;
-    }
-  return grub_elf64_load (elf, offset_phdr, 0, 0);
+  return grub_elf64_load (elf, filename, (void *) (grub_addr_t) (seg_addr - base_addr), GRUB_ELF_LOAD_FLAGS_62BITS, 0, 0);
 }
 
 static grub_err_t
@@ -253,7 +266,7 @@ grub_cmd_linux (grub_command_t cmd __attribute__ ((unused)),
 
   if (argc == 0)
     {
-      grub_error (GRUB_ERR_BAD_ARGUMENT, "no kernel specified");
+      grub_error (GRUB_ERR_BAD_ARGUMENT, N_("filename expected"));
       goto out;
     }
 
@@ -264,7 +277,7 @@ grub_cmd_linux (grub_command_t cmd __attribute__ ((unused)),
   if (elf->ehdr.ehdr32.e_type != ET_EXEC && elf->ehdr.ehdr32.e_type != ET_DYN)
     {
       grub_error (GRUB_ERR_UNKNOWN_OS,
-		  "this ELF file is not of the right type");
+		  N_("this ELF file is not of the right type"));
       goto out;
     }
 
@@ -272,13 +285,13 @@ grub_cmd_linux (grub_command_t cmd __attribute__ ((unused)),
   grub_loader_unset ();
 
   if (grub_elf_is_elf32 (elf))
-    grub_linux_load32 (elf);
+    grub_linux_load32 (elf, argv[0]);
   else
   if (grub_elf_is_elf64 (elf))
-    grub_linux_load64 (elf);
+    grub_linux_load64 (elf, argv[0]);
   else
     {
-      grub_error (GRUB_ERR_BAD_FILE_TYPE, "unknown ELF class");
+      grub_error (GRUB_ERR_BAD_FILE_TYPE, N_("invalid arch-dependent ELF magic"));
       goto out;
     }
 
@@ -317,30 +330,29 @@ static grub_err_t
 grub_cmd_initrd (grub_command_t cmd __attribute__ ((unused)),
 		 int argc, char *argv[])
 {
-  grub_file_t file = 0;
-  grub_ssize_t size;
+  grub_size_t size = 0;
   grub_addr_t first_addr;
   grub_addr_t addr;
+  struct grub_linux_initrd_context initrd_ctx = { 0, 0, 0 };
 
   if (argc == 0)
     {
-      grub_error (GRUB_ERR_BAD_ARGUMENT, "no initrd specified");
+      grub_error (GRUB_ERR_BAD_ARGUMENT, N_("filename expected"));
       goto fail;
     }
 
   if (!loaded)
     {
-      grub_error (GRUB_ERR_BAD_ARGUMENT, "you need to load the kernel first");
+      grub_error (GRUB_ERR_BAD_ARGUMENT, N_("you need to load the kernel first"));
       goto fail;
     }
 
-  grub_file_filter_disable_compression ();
-  file = grub_file_open (argv[0]);
-  if (! file)
+  if (grub_initrd_init (argc, argv, &initrd_ctx))
     goto fail;
 
+  size = grub_get_initrd_size (&initrd_ctx);
+
   first_addr = linux_addr + linux_size;
-  size = grub_file_size (file);
 
   /* Attempt to claim at a series of addresses until successful in
      the same way that grub_rescue_cmd_linux does.  */
@@ -350,19 +362,14 @@ grub_cmd_initrd (grub_command_t cmd __attribute__ ((unused)),
 
   grub_dprintf ("loader", "Loading initrd at 0x%x, size 0x%x\n", addr, size);
 
-  if (grub_file_read (file, (void *) addr, size) != size)
-    {
-      grub_ieee1275_release (addr, size);
-      grub_error (GRUB_ERR_FILE_READ_ERROR, "couldn't read file");
-      goto fail;
-    }
+  if (grub_initrd_load (&initrd_ctx, argv, (void *) addr))
+    goto fail;
 
   initrd_addr = addr;
   initrd_size = size;
 
  fail:
-  if (file)
-    grub_file_close (file);
+  grub_initrd_close (&initrd_ctx);
 
   return grub_errno;
 }
